"""
Building Block
==============

"""


import logging
import os
import numpy as np
import itertools as it
import rdkit.Chem.AllChem as rdkit
from glob import glob
from functools import partial
from scipy.spatial.distance import euclidean

from .. import elements
from ..elements import Atom
from .. import bonds
from ..bonds import Bond
from .molecule import Molecule
from ..functional_groups import fg_types
from ...utilities import vector_angle, dedupe, remake


logger = logging.getLogger(__name__)


class BuildingBlock(Molecule):
    """
    Represents a building block of a :class:`.ConstructedMolecule`.

    A :class:`BuildingBlock` can represent either an entire molecule or
    a molecular fragments used to construct a
    :class:`.ConstructedMolecule`. The building block uses
    :attr:`func_groups` to identify which atoms are modified during
    construction. Available functional group types can be seen in
    :attr:`functional_group_types`. Additional functional groups
    can be added at runtime by adding a :class:`.FGType` instance
    into :data:`stk.fg_types`. See :ref:`adding functional groups` for
    an example.

    Attributes
    ----------
    atoms : :class:`tuple` of :class:`.Atom`
        The atoms of the molecule.

    bonds : :class:`tuple` of :class:`.Bond`
        The bonds of the molecule.

    func_groups : :class:`tuple` of :class:`.FunctionalGroup`
        The functional groups present in the molecule. The
        id of a :class:`.FunctionalGroup` is its index.

    """

    # Put this here so that available functional groups appear in the
    # compiled documentation.
    functional_group_types = list(fg_types.keys())

    # Maps file extensions to functions which can be used to
    # create an rdkit molecule from that file type.
    _init_funcs = {
        '.mol': partial(
            rdkit.MolFromMolFile,
            sanitize=False,
            removeHs=False
        ),

        '.sdf': partial(
            rdkit.MolFromMolFile,
            sanitize=False,
            removeHs=False
        ),

        '.pdb': partial(
            rdkit.MolFromPDBFile,
            sanitize=False,
            removeHs=False
        ),
    }

    def __init__(
        self,
        smiles,
        functional_groups=None,
        random_seed=4,
        use_cache=False
    ):
        """
        Initialize from a SMILES string.

        Notes
        -----
        The molecule is given 3D coordinates with
        :func:`rdkit.ETKDGv2()`.

        Parameters
        ----------
        smiles : :class:`str`
            A SMILES string of the molecule.

        functional_groups : :class:`iterable` of :class:`str`, optional
            The names of the functional group types which are to be
            added to :attr:`func_groups`. If ``None``, then no
            functional groups are added.

        random_seed : :class:`int`, optional
            Random seed passed to :func:`rdkit.ETKDGv2`

        use_cache : :class:`bool`, optional
            If ``True``, a new :class:`.BuildingBlock` will
            not be made if a cached and identical one already exists,
            the one which already exists will be returned. If ``True``
            and a cached, identical :class:`BuildingBlock` does not
            yet exist the created one will be added to the cache.

        """

        # This method does not get called, See _construct().
        raise RuntimeError('This method should not be getting called.')

    @classmethod
    def _construct(
        cls,
        smiles,
        functional_groups=None,
        random_seed=4,
        use_cache=False
    ):
        obj = cls.__new__(cls)
        if functional_groups is None:
            functional_groups = ()

        mol = rdkit.AddHs(rdkit.MolFromSmiles(smiles))
        params = rdkit.ETKDGv2()
        params.randomSeed = random_seed
        for i in range(100):
            failed = rdkit.EmbedMolecule(mol, params) == -1
            if failed:
                params.randomSeed += 1
            else:
                break

        if params.randomSeed != random_seed:
            msg = (
                'Embedding with seed value of '
                f'"{random_seed}" failed. Using alternative value '
                f'of "{params.randomSeed}" was successful.'
            )
            logger.warning(msg)

        identity_key = cls._get_identity_key_from_rdkit_mol(
            mol=mol,
            functional_groups=functional_groups
        )
        if use_cache and identity_key in cls._cache:
            return cls._cache[identity_key]

        rdkit.Kekulize(mol)
        obj._init_from_rdkit_mol(
            mol=mol,
            functional_groups=functional_groups,
            identity_key=identity_key
        )
        if use_cache:
            cls._cache[identity_key] = obj
        return obj

    @classmethod
    def init_from_file(
        cls,
        path,
        functional_groups=None,
        use_cache=False
    ):
        """
        Initialize from a file.

        Parameters
        ----------
        path : :class:`str`
            The path to a molecular structure file. Supported file
            types are:

                #. ``.mol``, ``.sdf`` - MDL V3000 MOL file
                #. ``.pdb`` - PDB file

        functional_groups : :class:`iterable` of :class:`str`, optional
            The names of the functional group types which are to be
            added to :attr:`func_groups`. If ``None``, then no
            functional groups are added.

        use_cache : :class:`bool`, optional
            If ``True``, a new :class:`.BuildingBlock` will
            not be made if a cached and identical one already exists,
            the one which already exists will be returned. If ``True``
            and a cached, identical :class:`BuildingBlock` does not
            yet exist the created one will be added to the cache.

        Returns
        -------
        :class:`.BuildingBlock`
            The building block.

        Raises
        ------
        :class:`ValueError`
            If the file type cannot be used for initialization.

        """

        if os.path.exists(path):
            _, ext = os.path.splitext(path)

            if ext not in cls._init_funcs:
                raise ValueError(
                    f'Unable to initialize from "{ext}" files.'
                )
            # This remake needs to be here because molecules loaded
            # with rdkit often have issues, because rdkit tries to do
            # bits of structural analysis like stereocenters. remake
            # gets rid of all this problematic metadata.
            mol = remake(cls._init_funcs[ext](path))

        return cls.init_from_rdkit_mol(
            mol=mol,
            functional_groups=functional_groups,
            use_cache=use_cache
        )

    @classmethod
    def init_from_random_file(
        cls,
        file_glob,
        functional_groups=None,
        use_cache=False
    ):
        """
        Initialize from a random file in `file_glob`.

        Parameters
        ----------
        file_glob : :class:`str`
            A glob specifying files, one of which is used to initialize
            a :class:`.BuildingBlock` at random.

        functional_groups : :class:`iterable` of :class:`str`, optional
            The names of the functional group types which are to be
            added to :attr:`func_groups`. If ``None``, then no
            functional groups are added.

        use_cache : :class:`bool`, optional
            If ``True``, a new :class:`.BuildingBlock` will
            not be made if a cached and identical one already exists,
            the one which already exists will be returned. If ``True``
            and a cached, identical :class:`BuildingBlock` does not
            yet exist the created one will be added to the cache.

        Returns
        -------
        :class:`BuildingBlock`
            A random molecule from `file_glob`.

        Raises
        ------
        :class:`RuntimeError`
            If no files in `file_glob` could be initialized from.

        """

        files = glob(file_glob)
        np.random.shuffle(files)

        for path in files:
            try:
                return cls.init_from_file(
                    path=path,
                    functional_groups=functional_groups,
                    use_cache=use_cache
                )

            except Exception:
                msg = (
                    'Could not initialize '
                    f'{cls.__name__} from {path}.'
                )
                logger.warning(msg)
        raise RuntimeError(
            f'No files in "{file_glob}" could be initialized from.'
        )

    @classmethod
    def init_from_rdkit_mol(
        cls,
        mol,
        functional_groups=None,
        use_cache=False
    ):
        """
        Initialize from an :mod:`rdkit` molecule.

        Parameters
        ----------
        mol : :class:`rdkit.Mol`
            The molecule.

        functional_groups : :class:`iterable` of :class:`str`, optional
            The names of the functional group types which are to be
            added to :attr:`func_groups`. If ``None``, then no
            functional groups are added.

        use_cache : :class:`bool`, optional
            If ``True``, a new :class:`.BuildingBlock` will
            not be made if a cached and identical one already exists,
            the one which already exists will be returned. If ``True``
            and a cached, identical :class:`BuildingBlock` does not
            yet exist the created one will be added to the cache.

        Returns
        -------
        :class:`BuildingBlock`
            The molecule.

        """

        key = cls._get_identity_key_from_rdkit_mol(
            mol=mol,
            functional_groups=functional_groups
        )
        if use_cache and key in cls._cache:
            return cls._cache[key]

        bb = cls.__new__(cls)
        cls._init_from_rdkit_mol(
            self=bb,
            mol=mol,
            functional_groups=functional_groups,
            identity_key=key
        )

        if use_cache:
            cls._cache[key] = bb

        return bb

    def _init_from_rdkit_mol(
        self,
        mol,
        functional_groups,
        identity_key
    ):
        """
        Initialize from an :mod:`rdkit` molecule.

        Parameters
        ----------
        mol : :class:`rdkit.Mol`
            The molecule.

        functional_groups : :class:`iterable` of :class:`str`
            The names of the functional group types which are to be
            added to :attr:`func_groups`. If ``None`, then no
            functional groups are added.

        identity_key : :class:`tuple`
            The identity key of the molecule.

        Returns
        -------
        None : :class:`NoneType`

        """

        if functional_groups is None:
            functional_groups = ()

        atoms = tuple(
            Atom(a.GetIdx(), a.GetAtomicNum(), a.GetFormalCharge())
            for a in mol.GetAtoms()
        )
        bonds = tuple(
            Bond(
                atoms[b.GetBeginAtomIdx()],
                atoms[b.GetEndAtomIdx()],
                b.GetBondTypeAsDouble()
            )
            for b in mol.GetBonds()
        )
        position_matrix = mol.GetConformer().GetPositions()

        super().__init__(atoms, bonds, position_matrix, identity_key)

        fg_makers = (fg_types[name] for name in functional_groups)
        self.func_groups = tuple(
            func_group
            for fg_maker in fg_makers
            for func_group in fg_maker.get_functional_groups(self)
        )

    @classmethod
    def _init_from_dict(cls, mol_dict, use_cache):
        """
        Intialize from a :class:`dict` representation.

        Parameters
        ----------
        mol_dict : :class:`dict`
            A :class:`dict` representation of a molecule generated
            by :meth:`to_dict`.

        use_cache : :class:`bool`
            If ``True``, a new instance will not be made if a cached
            and identical one already exists, the one which already
            exists will be returned. If ``True`` and a cached,
            identical instance does not yet exist the created one will
            be added to the cache.

        Returns
        -------
        :class:`BuildingBlock`
            The molecule described by `mol_dict`.

        """

        d = dict(mol_dict)
        identity_key = eval(d.pop('identity_key'))
        if use_cache and identity_key in cls._cache:
            return cls._cache[identity_key]

        d.pop('class')
        functional_groups = d.pop('func_groups')

        obj = cls.__new__(cls)
        obj._identity_key = identity_key
        obj._position_matrix = np.array(d.pop('position_matrix')).T
        # If the cache is not being used, make sure to update all the
        # atoms and attributes to those in the dict.
        obj.atoms = eval(d.pop('atoms'), vars(elements))
        obj.bonds = eval(d.pop('bonds'), vars(bonds))
        for bond in obj.bonds:
            bond.atom1 = obj.atoms[bond.atom1]
            bond.atom2 = obj.atoms[bond.atom2]

        fg_makers = (fg_types[name] for name in functional_groups)
        obj.func_groups = tuple(
            func_group
            for fg_maker in fg_makers
            for func_group in fg_maker.get_functional_groups(obj)
        )
        for attr, val in d.items():
            setattr(obj, attr, eval(val))

        if use_cache:
            cls._cache[identity_key] = obj
        return obj

    def get_bonder_ids(self, fg_ids=None):
        """
        Yield ids of bonder atoms.

        Parameters
        ----------
        fg_ids : :class:`iterable` of :class:`int`
            The ids of functional groups whose bonder atoms should be
            yielded. If ``None`` then all bonder atom ids in the
            :class:`.BuildingBlock` will be yielded.

        Yields
        ------
        :class:`int`
            The id of a bonder atom.

        """

        if fg_ids is None:
            fg_ids = range(len(self.func_groups))

        for fg_id in fg_ids:
            yield from self.func_groups[fg_id].get_bonder_ids()

    def get_bonder_centroids(self, fg_ids=None):
        """
        Yield the centroids of bonder atoms.

        A bonder centroid is the centroid of all bonder atoms in a
        particular functional group.

        Parameters
        ----------
        fg_ids : :class:`iterable` of :class:`int`
            The ids of functional groups to be used. The bonder
            centroids will be yielded in this order.
            If ``None`` then all functional groups are used and
            centroids are yielded in ascending order of functional
            group id.

        Yields
        ------
        :class:`numpy.ndarray`
            The centroid of a functional group.

        """

        if fg_ids is None:
            fg_ids = range(len(self.func_groups))

        for fg_id in fg_ids:
            yield self.get_centroid(
                atom_ids=self.func_groups[fg_id].get_bonder_ids()
            )

    def get_bonder_plane(self, fg_ids=None):
        """
        Return coeffs of the plane formed by the bonder centroids.

        A bonder centroid is the centroid of all bonder atoms in a
        particular functional group.

        A plane is defined by the scalar plane equation::

            ax + by + cz = d.

        This method returns the ``a``, ``b``, ``c`` and ``d``
        coefficients of this equation for the plane formed by the
        bonder centroids. The coefficents ``a``, ``b`` and ``c``
        describe the normal vector to the plane. The coefficent ``d``
        is found by substituting these coefficients along with the
        x, y and z variables in the scalar equation and
        solving for ``d``. The variables x, y and z are
        substituted by the coordinates of some point on the plane. For
        example, the position of one of the bonder centroids.

        Parameters
        ----------
        fg_ids : :class:`iterable` of :class:`int`
            The ids of functional groups used to construct the plane.
            If there are more than three, a plane of best fit through
            the bonder centroids of the functional groups will be made.
            If ``None``, all functional groups in the
            :class:`BuildingBlock` will be used.

        Returns
        -------
        :class:`numpy.ndarray`
            This array has the form ``[a, b, c, d]`` and represents the
            scalar equation of the plane formed by the bonder
            centroids.

        References
        ----------
        https://tinyurl.com/okpqv6

        """

        if fg_ids is None:
            fg_ids = range(len(self.func_groups))
        else:
            # Iterable is used multiple times.
            fg_ids = list(fg_ids)

        centroid = self.get_centroid(
            atom_ids=self.func_groups[fg_ids[0]].get_bonder_ids()
        )
        normal = self.get_bonder_plane_normal(fg_ids=fg_ids)
        d = np.sum(normal * centroid)
        return np.append(normal, d)

    def get_bonder_plane_normal(self, fg_ids=None):
        """
        Return the normal to the plane formed by bonder centroids.

        A bonder centroid is the centroid of all bonder atoms in a
        particular functional group.

        The normal of the plane is defined such that it goes in the
        direction toward the centroid of the molecule.

        Parameters
        ----------
        fg_ids  : :class:`iterable` of :class:`int`, optional
            The ids of functional groups used to construct the plane.
            If there are more than three, a plane of best fit through
            the bonder centroids of the functional groups will be made.
            If ``None``, all functional groups in the
            :class:`BuildingBlock` will be used.

        Returns
        -------
        :class:`numpy.ndarray`
            A unit vector which describes the normal to the plane of
            the bonder centroids.

        Raises
        ------
        :class:`ValueError`
            If there are not at least 3 functional groups, which is
            necessary to define a plane.

        """

        if fg_ids is None:
            fg_ids = range(len(self.func_groups))
        else:
            # The iterable is used mutliple times.
            fg_ids = list(fg_ids)

        if len(fg_ids) < 3:
            raise ValueError(
                'At least 3 functional groups '
                'are necessary to create a plane.'
            )

        centroids = np.array(list(self.get_bonder_centroids(
            fg_ids=fg_ids
        )))
        bonder_centroid = self.get_centroid(
            atom_ids=self.get_bonder_ids(fg_ids=fg_ids)
        )
        normal = np.linalg.svd(centroids - bonder_centroid)[-1][2, :]
        cc_vector = self.get_centroid_centroid_direction_vector(
            fg_ids=fg_ids
        )
        if (
            # vector_angle is NaN if cc_vector is [0, 0, 0].
            not np.allclose(cc_vector, [0, 0, 0], atol=1e-5)
            and vector_angle(normal, cc_vector) > np.pi/2
        ):
            normal *= -1
        return normal

    def get_bonder_distances(self, fg_ids=None):
        """
        Yield distances between pairs of bonder centroids.

        A bonder centroid is the centroid of all bonder atoms in a
        particular functional group.

        Parameters
        ----------
        fg_ids : :class:`iterable` of :class:`int`
            The ids of functional groups to be used.
            If ``None`` then all functional groups are used.

        Yields
        ------
        :class:`tuple`
            A :class:`tuple` of the form ``(3, 54, 12.54)``.
            The first two elements are the ids of the involved
            functional groups and the third element is the distance
            between them.

        """

        if fg_ids is None:
            fg_ids = range(len(self.func_groups))

        # Iterator yielding tuples of form (fg_id, bonder_centroid)
        centroids = ((
            i, self.get_centroid(
                atom_ids=self.func_groups[i].get_bonder_ids()
            ))
            for i in fg_ids
        )
        pairs = it.combinations(iterable=centroids, r=2)
        for (id1, c1), (id2, c2) in pairs:
            yield id1, id2, float(euclidean(c1, c2))

    def get_bonder_direction_vectors(
        self,
        fg_ids=None
    ):
        """
        Yield the direction vectors between bonder centroids.

        A bonder centroid is the centroid of all bonder atoms in a
        particular functional group.

        Parameters
        ----------
        fg_ids : :class:`iterable` of :class:`int`
            The ids of functional groups to be used.
            If ``None`` then all functional groups are used.

        Yields
        ------
        :class:`tuple`
            They yielded tuple has the form

            .. code-block:: python

                (3, 54, np.array([12.2, 43.3, 9.78]))

            The first two elements of the tuple represent the ids
            of the start and end fgs of the vector, respectively. The
            array is the direction vector running between the
            functional group positions.

        """

        if fg_ids is None:
            fg_ids = range(len(self.func_groups))

        # Iterator yielding tuples of form (fg_id, bonder_centroid)
        centroids = ((
            i, self.get_centroid(
                atom_ids=self.func_groups[i].get_bonder_ids()
            ))
            for i in fg_ids
        )
        pairs = it.combinations(iterable=centroids, r=2)
        for (id1, c1), (id2, c2) in pairs:
            yield id2, id1, c1-c2

    def get_centroid_centroid_direction_vector(
        self,
        fg_ids=None
    ):
        """
        Return the direction vector between the 2 molecular centroids.

        The first molecular centroid is the centroid of the entire
        molecule. The second molecular centroid is the of the
        bonder atoms in the molecule.

        Parameters
        ----------
        fg_ids : :class:`iterable` of :class:`int`
            The ids of functional groups to be used for calculating the
            bonder centroid. If ``None`` then all functional groups are
            used.

        Returns
        -------
        :class:`numpy.ndarray`
            The vector running from the centroid of the bonder atoms to
            the molecular centroid.

        """

        if fg_ids is None:
            fg_ids = range(len(self.func_groups))
        else:
            # This iterable gets used more than once.
            fg_ids = list(fg_ids)

        bonder_centroid = self.get_centroid(
            atom_ids=self.get_bonder_ids(fg_ids=fg_ids)
        )
        return self.get_centroid() - bonder_centroid

    def to_dict(self, include_attrs=None, ignore_missing_attrs=False):
        """
        Return a :class:`dict` representation.

        Parameters
        ----------
        include_attrs : :class:`list` of :class:`str`, optional
            The names of additional attributes of the molecule to be
            added to the :class:`dict`. Each attribute is saved as a
            string using :func:`repr`.

        ignore_missing_attrs : :class:`bool`, optional
            If ``False`` and an attribute in `include_attrs` is not
            held by the :class:`BuildingBlock`, an error will be
            raised.

        Returns
        -------
        :class:`dict`
            A :class:`dict` which represents the molecule.

        """

        if include_attrs is None:
            include_attrs = []

        fgs = list(dedupe(fg.fg_type.name for fg in self.func_groups))

        bonds = []
        for bond in self.bonds:
            clone = bond.clone()
            clone.atom1 = clone.atom1.id
            clone.atom2 = clone.atom2.id
            bonds.append(clone)

        d = {
            'class': self.__class__.__name__,
            'func_groups': fgs,
            'position_matrix': self.get_position_matrix().tolist(),
            'atoms': repr(self.atoms),
            'bonds': repr(bonds),
            'identity_key': repr(self._identity_key)
        }

        if ignore_missing_attrs:
            d.update({
                attr: repr(getattr(self, attr))
                for attr in include_attrs
                if hasattr(self, attr)
            })
        else:
            d.update({
                attr: repr(getattr(self, attr))
                for attr in include_attrs
            })

        return d

    @staticmethod
    def _get_identity_key_from_rdkit_mol(mol, functional_groups):
        if functional_groups is None:
            functional_groups = ()
        functional_groups = sorted(functional_groups)

        # Don't modify the original molecule.
        mol = rdkit.Mol(mol)
        rdkit.SanitizeMol(mol)
        return (
            *functional_groups,
            rdkit.MolToSmiles(mol, canonical=True)
        )

    def __str__(self):
        smiles = rdkit.MolToSmiles(rdkit.RemoveHs(self.to_rdkit_mol()))
        func_groups = list(dedupe(
            fg.fg_type.name for fg in self.func_groups
        ))
        return f'{self.__class__.__name__}({smiles!r}, {func_groups})'

    def __repr__(self):
        return str(self)



"""
Defines MacroModel optimizers.

"""

import os
import subprocess as sp
import time
import rdkit.Chem.AllChem as rdkit
import psutil
import re
from uuid import uuid4
import logging
import gzip

from ...utilities import MAEExtractor, move_generated_macromodel_files
from .optimizers import Optimizer

logger = logging.getLogger(__name__)


class MacroModelConversionError(Exception):
    ...


class MacroModelPathError(Exception):
    ...


class MacroModelForceFieldError(Exception):
    ...


class MacroModelOptimizationError(Exception):
    ...


class MacroModelLewisStructureError(Exception):
    ...


class MacroModelInputError(Exception):
    ...


class _MacroModel(Optimizer):
    """
    Base class for MacroModel optimzers.

    """

    def __init__(
        self,
        macromodel_path,
        output_dir,
        timeout,
        force_field,
        maximum_iterations,
        minimum_gradient,
        use_cache
    ):
        """
        Initialize a :class:`_MacroModel` instance.

        Parameters
        ----------
        macromodel_path : :class:`str`
            The full path of the Schrodinger suite within the user's
            machine. For example, on a Linux machine this may be
            something like ``'/opt/schrodinger2017-2'``.

        output_dir : :class:`str`
            The name of the directory into which files generated during
            the optimization are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        timeout : :class:`float`
            The amount in seconds the optimization is allowed to run
            before being terminated. ``None`` means there is no
            timeout.

        force_field : :class:`int`
            The number of the force field to be used.

        maximum_iterations : :class:`int`
            The maximum number of iterations done during the
            optimization. Cannot be more than ``999999``.

        minimum_gradient : :class:`float`
            The gradient at which optimization is stopped.
            Cannot be less than ``0.0001``.

        use_cache : :class:`bool`
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule.

        """

        self._macromodel_path = macromodel_path
        self._output_dir = output_dir
        self._timeout = timeout
        self._force_field = force_field
        self._maximum_iterations = maximum_iterations
        self._minimum_gradient = minimum_gradient
        super().__init__(use_cache=use_cache)

    def _run_bmin(self, mol, run_name):
        """
        Run an optimization using bmin.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule being optimized.

        run_name : :class:`str`
            The name of the run. The files generated by this run
            will have this name.

        Returns
        -------
        None : :class:`NoneType`

        Raises
        ------
        :class:`MacroModelOptimizationError`
            If the optimization failed for some unspecified.

        :class:`MacroModelForceFieldError`
            If the force field could not be used with the molecule.

        :class:`MacroModelLewisStructureError`
            If Lewis structure of the molecule had issues.

        :class:`MacroModelPathError`
            If an invalid MacroModel path is being used.

        """

        logger.info(f'Running bmin on "{mol}".')

        # To run MacroModel a command is issued to the console via
        # ``subprocess.Popen``. The command is the full path of the
        # ``bmin`` program. ``bmin`` is located in the Schrodinger
        # installation folder.
        log_file = f'{run_name}.log'
        opt_app = os.path.join(self._macromodel_path, 'bmin')
        # The first member of the list is the command, the following
        # ones are any additional arguments.

        opt_cmd = [opt_app, run_name, '-WAIT', '-LOCAL']

        incomplete = True
        while incomplete:
            process = psutil.Popen(
                opt_cmd,
                stdout=sp.PIPE,
                stderr=sp.STDOUT,
                universal_newlines=True
            )
            try:
                output, _ = process.communicate(timeout=self._timeout)

            except sp.TimeoutExpired:
                logger.warning(
                    'Minimization took too long and was terminated '
                    f'by force on "{mol}".'
                )
                self._kill_bmin(mol)
                output = ''

            logger.debug(
                f'Output of bmin on "{mol}" was: {output}.'
            )

            with open(log_file, 'r') as log:
                log_content = log.read()

            # Check the log for error reports.
            error1 = 'termination due to error condition           21-'
            if error1 in log_content:
                raise MacroModelOptimizationError(
                    'bmin crashed. See log file.'
                )

            error2 = 'FATAL do_nosort_typing: NO MATCH found for atom '
            if error2 in log_content:
                raise MacroModelForceFieldError(
                    'The log implies the force field failed.'
                )

            error3 = (
                'FATAL gen_lewis_structure(): '
                'could not find best Lewis structure'
            )
            error4 = (
                'skipping input structure  '
                'due to forcefield interaction errors'
            )
            if error3 in log_content and error4 in log_content:
                raise MacroModelLewisStructureError(
                    'bmin failed due to poor Lewis structure.'
                )

            if 'MDYN error encountered' in log_content:
                raise MacroModelOptimizationError(
                    'MD error during optimization.'
                )

            # If optimization fails because a wrong Schrodinger path
            # was given, raise.
            if 'The system cannot find the path specified' in output:
                raise MacroModelPathError(
                    'Invalid Schrodinger path given to bmin.'
                )

            # If optimization fails because the license is not found,
            # rerun the function.
            if self._license_found(run_name, output, mol):
                incomplete = False

        # Make sure the .maegz file created by the optimization is
        # present.
        maegz = f'{run_name}-out.maegz'
        self._wait_for_file(maegz)
        if not os.path.exists(log_file) or not os.path.exists(maegz):
            raise MacroModelOptimizationError(
                'The .log and/or .maegz files were not created.'
            )

    def _kill_bmin(self, mol, run_name):
        """
        Kill bmin.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule being optimized.

        run_name : :class:`str`
            The name of the run. The files generated by this run will
            have this name.

        Returns
        -------
        None : :class:`NoneType`

        """

        name = re.split(r'\\|/', run_name)[-1]
        app = os.path.join(self._macromodel_path, 'jobcontrol')
        cmd = [app, '-stop', name]

        incomplete = True
        while incomplete:
            out = sp.run(
                cmd,
                stdout=sp.PIPE,
                stderr=sp.STDOUT,
                universal_newlines=True
            )

            # Keep re-running the function until license is found.
            if self._license_found(run_name, out.stdout):
                incomplete = False

        # This loop causes the function to wait until the job has been
        # killed via job control. This means the output files will have
        # been written by the time the function exits. Essentially the
        # loop continues until the job is no longer found by
        # "./jobcontrol -list"
        cmd = [app, '-list']
        output = name
        start = time.time()
        while name in output:
            output = sp.run(
                cmd,
                stdout=sp.PIPE,
                stderr=sp.STDOUT,
                universal_newlines=True
            ).stdout
            if time.time() - start > 600:
                break

    def _license_found(self, run_name, output, mol=None):
        """
        Check to see if minimization failed due to a missing license.

        The user can be notified of this in one of two ways. Sometimes
        the output of the submission contains the message informing
        that the license was not found and in other cases it will be
        the log file. This function checks both of these sources for
        this message.

        Parameters
        ----------
        run_name : :class:`str`
            The name of the run. The files generated by this run will
            have this name.

        output : :class:`str`
            The output from submitting the minimization of the
            structure to bmin.

        mol : :class:`.Molecule`, optional
            The molecule being optimized. If the ``.log`` file is not
            to be checked, the default ``None`` should be used.

        Returns
        -------
        :class:`bool`
            ``True`` if the license was found. ``False`` if the
            minimization did not occur due to a missing license.

        """

        if 'Could not check out a license for mmlibs' in output:
            return False
        if mol is None:
            return True

        # To check if the log file mentions a missing license file open
        # the log file and scan for the apporpriate string.

        # Check if the file exists first. If not, this is often means
        # the calculation must be redone so return False anyway.
        with open(f'{run_name}.log', 'r') as log_file:
            log_file = log_file.read()

        if 'Could not check out a license for mmlibs' in log_file:
            return False

        return True

    @staticmethod
    def _get_com_line(
        arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9
    ):
        return (
            f' {arg1:<5}{arg2:>7}{arg3:>7}'
            f'{arg4:>7}{arg5:>7}{arg6:>11.4f}'
            f'{arg7:>11.4f}{arg8:>11.4f}{arg9:>11.4f}'
        )

    def _run_structconvert(self, input_path, output_path):
        """
        Use structconvert to change file type.

        Parameters
        ----------
        input_path : :class:`str`
            The path to the input file.

        output_path : :class:`str`
            The path to the output file.

        Returns
        -------
        None : :class:`NoneType`

        """

        convrt_app = os.path.join(
            self._macromodel_path, 'utilities', 'structconvert'
        )
        convrt_cmd = [convrt_app, input_path, output_path]

        incomplete = True
        while incomplete:

            # Execute the file conversion.
            try:
                convrt_return = sp.run(
                    convrt_cmd,
                    stdout=sp.PIPE,
                    stderr=sp.STDOUT,
                    universal_newlines=True
                )

            # If conversion fails because a wrong Schrodinger path was
            # given, raise.
            except FileNotFoundError:
                raise MacroModelPathError(
                    'Wrong Schrodinger path supplied to structconvert.'
                )

            if 'File does not exist' in convrt_return.stdout:
                raise MacroModelConversionError(
                    f'structconvert input file, {input_path}, '
                    f'missing. Console output was '
                    f'{convrt_return.stdout}'
                )

            # Keep re-running the function until license is found.
            run_name, _ = os.path.splitext(input_path)
            if self._license_found(run_name, convrt_return.stdout):
                incomplete = False

        # If force field failed, raise.
        if 'number 1' in convrt_return.stdout:
            raise MacroModelForceFieldError(convrt_return.stdout)

        self._wait_for_file(output_path)
        if not os.path.exists(output_path):
            raise MacroModelConversionError(
                f'Conversion output file {output_path} was not found.'
                f' Console output was {convrt_return.stdout}.'
            )

        return convrt_return

    def _wait_for_file(self, path, timeout=10):
        """
        Wait until a given file exists or `timeout` expires.

        Parameters
        ----------
        path : :class:`str`
            The full path of the file which should be waited for.

        timeout : :class:`int` or :class:`float`, optional
            The number of seconds before the function stops waiting and
            returns.

        Returns
        --------
        None : :class:`NoneType`

        """

        t_start = time.time()
        tick = 0
        while True:
            time_taken = time.time() - t_start
            if divmod(time_taken, 5)[0] == tick + 1:
                logger.warning(f'Waiting for "{path}".')
                tick += 1

            if os.path.exists(path) or time_taken > timeout:
                break

    def _convert_maegz_to_mae(self, run_name):
        """
        Convert a ``.maegz`` file to a ``.mae`` file.

        Parameters
        ----------
        run_name : :class:`str`
            The name of the run. The files generated by this run will
            have this name.

        Returns
        -------
        None : :class:`NoneType`

        """

        gz_file = gzip.open(f'{run_name}-out.maegz')
        with open(f'{run_name}.mae', 'wb') as f:
            f.write(gz_file.read())
        gz_file.close()

    def _fix_params(self, mol, com):
        """
        Fix bond distances and angles in ``.com`` file.

        For each bond distance, bond angle and torisional angle that
        does not involve a bond created by
        :meth:`~.Topology.construct`, a "FX" command is added to the
        body of the ``.com`` file.

        These lines replace the filler line in the main string.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule which is to be optimized.

        com : :class:`str`
            The body of the ``.com`` file which is to have fix commands
            added.

        Returns
        -------
        :class:`str`
            A string holding the body of the ``.com`` file with
            instructions to fix the various bond distances and angles
            as described in the docstring.

        """

        fix_block = ''
        # Add lines that fix the bond distance.
        fix_block = self._fix_distances(mol, fix_block)
        # Add lines that fix the bond angles.
        fix_block = self._fix_bond_angles(mol, fix_block)
        # Add lines that fix the torsional angles.
        fix_block = self._fix_torsional_angles(mol, fix_block)

        return com.replace(
            '!!!BLOCK_OF_FIXED_PARAMETERS_COMES_HERE!!!\n',
            fix_block
        )


class MacroModelForceField(_MacroModel):
    """
    Uses MacroModel force fields to optimize molecules.

    """

    def __init__(
        self,
        macromodel_path,
        output_dir=None,
        restricted=False,
        timeout=None,
        force_field=16,
        maximum_iterations=2500,
        minimum_gradient=0.05,
        use_cache=False
    ):
        """
        Initialize a :class:`MacroModelForceField` object.

        Parameters
        ----------
        macromodel_path : :class:`str`
            The full path of the Schrodinger suite within the user's
            machine. For example, on a Linux machine this may be
            something like ``'/opt/schrodinger2017-2'``.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the optimization are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        restricted : :class:`bool`, optional
            If ``True`` then an optimization is performed only on the
            bonds added by :meth:`~.Topology.construct`. If ``False``
            then all bonds are optimized.

        timeout : :class:`float`, optional
            The amount in seconds the optimization is allowed to run
            before being terminated. ``None`` means there is no
            timeout.

        force_field : :class:`int`, optional
            The number of the force field to be used.

        maximum_iterations : :class:`int`, optional
            The maximum number of iterations done during the
            optimization. Cannot be more than ``999999``.

        minimum_gradient : :class:`float`, optional
            The gradient at which optimization is stopped.
            Cannot be less than ``0.0001``.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule.

        """
        self._check_params(
            minimum_gradient=minimum_gradient,
            maximum_iterations=maximum_iterations
        )
        self._restricted = restricted
        super().__init__(
            macromodel_path=macromodel_path,
            output_dir=output_dir,
            force_field=force_field,
            maximum_iterations=maximum_iterations,
            minimum_gradient=minimum_gradient,
            timeout=timeout,
            use_cache=use_cache
        )

    @staticmethod
    def _check_params(minimum_gradient, maximum_iterations):
        """
        Check if the optimization parameters are valid for MacroModel.

        Parameters
        ----------
        minimum_gradient : :class:`float`
            The gradient at which optimization is stopped.
            Cannot be less than ``0.0001``.

        maximum_iterations : :class:`int`
            The maximum number of iterations done during the
            optimization. Cannot be more than ``999999``.

        Returns
        -------
        None : :class:`NoneType`

        Raises
        ------
        :class:`.MacroModelInputError`
            If the parameters cannot be converted into a valid ``.com``
            file entry.

        """

        if minimum_gradient < 0.0001:
            raise MacroModelInputError(
                'Convergence gradient (< 0.0001) is too small.'
            )

        if maximum_iterations > 999999:
            raise MacroModelInputError(
                'Number of iterations (> 999999) is too high.'
            )

    def _generate_com(self, mol, run_name):
        """
        Create a ``.com`` file for a MacroModel optimization.

        The created ``.com`` file fixes all bond parameters which were
        not added by :meth:`~.Topology.construct`. This means all bond
        distances, bond angles and torsional angles are fixed, except
        for cases where it involves a bond added by
        :meth:`.Topology.construct`.

        This fixing is implemented by creating a ``.com`` file with
        various "FX" commands written within its body.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule which is to be optimized.

        run_name : :class:`str`
            The name of the run. The files generated by this run will
            have this name.

        Returns
        -------
        None : :class:`NoneType`

        """

        logger.debug(f'Creating .com file for "{mol}".')

        # This is the body of the ``.com`` file. The line that begins
        # and ends with exclamation lines is replaced with the various
        # commands that fix bond distances and angles.
        line1 = ('FFLD', self._force_field, 1, 0, 0, 1, 0, 0, 0)
        line2 = ('BGIN', 0, 0, 0, 0, 0, 0, 0, 0)
        line3 = ('READ', 0, 0, 0, 0, 0, 0, 0, 0)
        line4 = ('CONV', 2, 0, 0, 0, self._minimum_gradient, 0, 0, 0)
        line5 = ('MINI', 1, 0, self._maximum_iterations, 0, 0, 0, 0, 0)
        line6 = ('END', 0, 1, 0, 0, 0, 0, 0, 0)

        com_block = "\n".join([
            self._get_com_line(*line1),
            self._get_com_line(*line2),
            self._get_com_line(*line3),
            '!!!BLOCK_OF_FIXED_PARAMETERS_COMES_HERE!!!',
            self._get_com_line(*line4),
            self._get_com_line(*line5),
            self._get_com_line(*line6)
        ])

        # If `restricted` is ``False`` do not add a fix block.
        if not self._restricted:
            com_block = com_block.replace(
                "!!!BLOCK_OF_FIXED_PARAMETERS_COMES_HERE!!!\n",
                ''
            )
        else:
            # This function adds all the lines which fix bond distances
            # and angles into com_block.
            com_block = self._fix_params(mol, com_block)

        # Writes the .com file.
        with open(f'{run_name}.com', 'w') as com:
            # The first line holds the .mae file containing the
            # molecule to be optimized.
            com.write(f'{run_name}.mae\n')
            # The second line holds the name of the output file of the
            # optimization.
            com.write(f'{run_name}-out.maegz\n')
            # Next is the body of the .com file.
            com.write(com_block)

    def optimize(self, mol):
        """
        Optimize a molecule.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        run_name = str(uuid4().int)
        if self._output_dir is None:
            output_dir = run_name
        else:
            output_dir = self._output_dir

        mol_path = f'{run_name}.mol'
        mae_path = f'{run_name}.mae'
        # First write a .mol file of the molecule.
        mol.write(mol_path)
        # MacroModel requires a ``.mae`` file as input.
        self._run_structconvert(mol_path, mae_path)
        # generate the ``.com`` file for the MacroModel run.
        self._generate_com(mol, run_name)
        # Run the optimization.
        self._run_bmin(mol, run_name)
        # Get the ``.maegz`` optimization output to a ``.mae``.
        self._convert_maegz_to_mae(run_name)
        mol.update_from_file(mae_path)
        move_generated_macromodel_files(run_name, output_dir)

    def _fix_distances(self, mol, fix_block):
        """
        Add lines fixing bond distances to ``.com`` body.

        Only bond distances which do not involve bonds created during
        construction are fixed.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        fix_block : :class:`str`
            A string holding fix commands in the ``.com`` file.

        Returns
        -------
        :class:`str`
            A string holding fix commands in the ``.com`` file.

        """

        bonder_ids = set(
            id_
            for fg in mol.func_groups
            for id_ in fg.get_bonder_ids()
        )

        # Go through all the bonds in the rdkit molecule. If the bond
        # is not between bonder atoms add a fix line to the
        # ``fix_block``. If the bond does invovle two bonder atoms go
        # to the next bond. This is because a bond between 2 bonder
        # atoms was added during construction and should therefore not
        # be fixed.
        for bond in mol.bonds:

            if (
                bond.atom1.id in bonder_ids
                and bond.atom2.id in bonder_ids
            ):
                continue

            # Make sure that the indices are increased by 1 in the .mae
            # file.
            atom1_id = bond.atom1.id + 1
            atom2_id = bond.atom2.id + 1
            args = ('FXDI', atom1_id, atom2_id, 0, 0, 99999, 0, 0, 0)
            fix_block += self._get_com_line(*args)
            fix_block += '\n'

        return fix_block

    def _fix_bond_angles(self, mol, fix_block):
        """
        Add lines fixing bond angles to the ``.com`` body.

        All bond angles of the molecule are fixed.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        fix_block : :class:`str`
            A string holding fix commands in the ``.com`` file.

        Returns
        -------
        :class:`str`
            A string holding fix commands in the ``.com`` file.

        """

        paths = rdkit.FindAllPathsOfLengthN(
            mol=mol.to_rdkit_mol(),
            length=3,
            useBonds=False,
            useHs=True
        )
        for atom_ids in paths:
            atom_ids = [i+1 for i in atom_ids]
            args = ('FXBA', *atom_ids, 99999, 0, 0, 0, 0)
            fix_block += self._get_com_line(*args)
            fix_block += '\n'

        return fix_block

    def _fix_torsional_angles(self, mol, fix_block):
        """
        Add lines fixing torsional bond angles to the ``.com`` body.

        All torsional angles of the molecule are fixed.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        fix_block : :class:`str`
            A string holding fix commands in the ``.com`` file.

        Returns
        -------
        :class:`str`
            A string holding fix commands in the ``.com`` file.

        """

        paths = rdkit.FindAllPathsOfLengthN(
            mol=mol.to_rdkit_mol(),
            length=4,
            useBonds=False,
            useHs=True
        )
        for atom_ids in paths:
            atom_ids = [i+1 for i in atom_ids]
            args = ('FXTA', *atom_ids, 99999, 361, 0, 0)
            fix_block += self._get_com_line(*args)
            fix_block += '\n'

        return fix_block


class MacroModelMD(_MacroModel):
    """
    Runs a molecular dynamics conformer search using MacroModel.

    """

    def __init__(
        self,
        macromodel_path,
        output_dir=None,
        timeout=None,
        force_field=16,
        temperature=750,
        conformers=50,
        time_step=1,
        eq_time=10,
        simulation_time=200,
        maximum_iterations=2500,
        minimum_gradient=0.05,
        restricted_bonds=None,
        restricted_bond_angles=None,
        restricted_torsional_angles=None,
        use_cache=False
    ):
        """
        Initialize a :class:`.MacroModelMD` instance.

        Parameters
        ----------
        macromodel_path : :class:`str`
            The full path of the Schrodinger suite within the user's
            machine. For example, on a Linux machine this may be
            something like ``'/opt/schrodinger2017-2'``.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the optimization are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        timeout : :class:`float`, optional
            The amount in seconds the MD is allowed to run before
            being terminated. ``None`` means there is no timeout.

        force_field : :class:`int`, optional
            The number of the force field to be used.

        temperature : :class:`float`, optional
            The temperature in Kelvin at which the MD is run.
            Cannot be more than ``99999.99``.

        conformers' : :class:`int`, optional
            The number of conformers sampled and optimized from the MD.
            Cannot be more than ``9999``.

        simulation_time : :class:`float`, optional
            The simulation time in ``ps`` of the MD.
            Cannot be more than ``999999.99``.

        time_step : :class:`float`, optional
            The time step in ``fs`` for the MD.
            Cannot be more than ``99999.99``.

        eq_time : :class:`float`, optional
            The equilibriation time in ``ps`` before the MD is run.
            Cannot be more than ``999999.99``.

        maximum_iterations : :class:`int`, optional
            The maximum number of iterations done during the
            optimization. Cannot be more than ``999999``.

        minimum_gradient : :class:`float`, optional
            The gradient at which optimization is stopped.
            Cannot be less than ``0.0001``.

        restricted_bonds : :class:`set`, optional
            A :class:`set` of the form

            .. code-block:: python

                restricted_bonds = {
                    frozenset((0, 10)),
                    frozenset((3, 14)),
                    frozenset((5, 6))
                }

            Where each :class:`frozenset` defines which bonds should
            have a fixed length via the atom ids of atoms in the bond.

        restricted_bond_angles : :class:`set`, optional
            A :class:`set` of the form

            .. code-block:: python

                restricted_bonds = {
                    frozenset((0, 10, 12)),
                    frozenset((3, 14, 7)),
                    frozenset((5, 8, 2))
                }

            Where each :class:`frozenset` defines which bond angles
            should have a fixed size via the atom ids of atoms in the
            bond angle.

        restricted_torsional_angles : :class:`set`, optional
            A :class:`set` of the form

            .. code-block:: python

                restricted_bonds = {
                    frozenset((0, 10, 12, 3)),
                    frozenset((3, 14, 7, 4)),
                    frozenset((5, 8, 2, 9))
                }

            Where each :class:`frozenset` defines which torsional
            angles should have a fixed size via the atom ids of atoms
            in the torsional angle.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule.

        """

        if restricted_bonds is None:
            restricted_bonds = set()
        if restricted_bond_angles is None:
            restricted_bond_angles = set()
        if restricted_torsional_angles is None:
            restricted_torsional_angles = set()

        self._check_params(
            temperature=temperature,
            conformers=conformers,
            simulation_time=simulation_time,
            time_step=time_step,
            eq_time=eq_time,
            minimum_gradient=minimum_gradient,
            maximum_iterations=maximum_iterations
        )

        self._temperature = temperature
        self._conformers = conformers
        self._time_step = time_step
        self._eq_time = eq_time
        self._simulation_time = simulation_time
        self._restricted_bonds = restricted_bonds
        self._restricted_bond_angles = restricted_bond_angles
        self._restricted_torsional_angles = restricted_torsional_angles

        # Negative simulation time is interpreted as times 100 ps.
        if simulation_time > 99999.99:
            self._sim_time = -simulation_time/100
        else:
            self._sim_time = simulation_time

        # Negative equilibration time is interpreted as times 100 ps.
        if eq_time > 99999.99:
            self._eq_time = -eq_time/100
        else:
            self._eq_time = eq_time

        super().__init__(
            macromodel_path=macromodel_path,
            output_dir=output_dir,
            timeout=timeout,
            force_field=force_field,
            maximum_iterations=maximum_iterations,
            minimum_gradient=minimum_gradient,
            use_cache=use_cache
        )

    @staticmethod
    def _check_params(
        temperature,
        conformers,
        simulation_time,
        time_step,
        eq_time,
        minimum_gradient,
        maximum_iterations
    ):
        """
        Check if the optimization parameters are valid for MacroModel.

        Parameters
        ----------
        temperature : :class:`float`
            The temperature in Kelvin at which the MD is run.
            Cannot be more than ``99999.99``.

        conformers' : :class:`int`
            The number of conformers sampled and optimized from the MD.
            Cannot be more than ``9999``.

        simulation_time : :class:`float`
            The simulation time in ``ps`` of the MD.
            Cannot be more than ``999999.99``.

        time_step : :class:`float`
            The time step in ``fs`` for the MD.
            Cannot be more than ``99999.99``.

        eq_time : :class:`float`
            The equilibriation time in ``ps`` before the MD is run.
            Cannot be more than ``999999.99``.

        minimum_gradient : :class:`float`
            The gradient at which optimization is stopped.
            Cannot be less than ``0.0001``.

        maximum_iterations : :class:`int`
            The maximum number of iterations done during the
            optimization. Cannot be more than ``999999``.

        Returns
        -------
        None : :class:`NoneType`

        Raises
        ------
        :class:`.MacroModelInputError`
            If the parameters cannot be converted into a valid ``.com``
            file entry.

        """

        if temperature > 99999.99:
            raise MacroModelInputError(
                'Supplied temperature (> 99999 K) is too high.'
            )

        if conformers > 9999:
            raise MacroModelInputError(
                'Supplied number of conformers (> 9999) is too high.'
            )

        if simulation_time > 999999.99:
            raise MacroModelInputError(
                'Supplied simulation time (> 999999 ps) is too long.'
            )

        if time_step > 99999.99:
            raise MacroModelInputError(
                'Supplied time step (> 99999 fs) is too high.'
            )

        if eq_time > 999999.99:
            raise MacroModelInputError(
                'Supplied eq time (> 999999 ps) is too long.'
            )

        if minimum_gradient < 0.0001:
            raise MacroModelInputError(
                'Convergence gradient (< 0.0001) is too small.'
            )

        if maximum_iterations > 999999:
            raise MacroModelInputError(
                'Number of iterations (> 999999) is too high.'
            )

    def _generate_com(self, mol, run_name):
        """
        Create a ``.com`` file for a MacroModel optimization.

        The created ``.com`` file fixes all bond parameters which were
        not added by :meth:`~.Topology.construct`. This means all bond
        distances, bond angles and torsional angles are fixed, except
        for cases where it involves a bond added by
        :meth:`~.Topology.construct`.

        This fixing is implemented by creating a ``.com`` file with
        various "FX" commands written within its body.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule which is to be optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        logger.debug(f'Creating .com file for "{mol}".')

        # Define some short aliases to keep the following lines neat.
        temp = self._temperature
        sim_time = self._sim_time
        tstep = self._time_step
        eq_time = self._eq_time

        line1 = ('FFLD', self._force_field, 1, 0, 0, 1, 0, 0, 0)
        line2 = ('READ', 0, 0, 0, 0, 0, 0, 0, 0)
        line3 = ('MDIT', 0, 0, 0, 0, temp, 0, 0, 0)
        line4 = ('MDYN', 0, 0, 0, 0, tstep, eq_time, temp, 0)
        line5 = ('MDSA', self._conformers, 0, 0, 0, 0, 0, 1, 0)
        line6 = ('MDYN', 1, 0, 0, 0, tstep, sim_time, temp, 0)
        line7 = ('WRIT', 0, 0, 0, 0, 0, 0, 0, 0)
        line8 = ('RWND', 0, 1, 0, 0, 0, 0, 0, 0)
        line9 = ('BGIN', 0, 0, 0, 0, 0, 0, 0, 0)
        line10 = ('READ', -2, 0, 0, 0, 0, 0, 0, 0)
        line11 = ('CONV', 2, 0, 0, 0, self._minimum_gradient, 0, 0, 0)
        line12 = ('MINI', 1, 0, self._maximum_iterations, 0, 0, 0, 0, 0)
        line13 = ('END', 0, 1, 0, 0, 0, 0, 0, 0)

        com_block = "\n".join([
            self._get_com_line(*line1),
            self._get_com_line(*line2),
            '!!!BLOCK_OF_FIXED_PARAMETERS_COMES_HERE!!!',
            self._get_com_line(*line3),
            self._get_com_line(*line4),
            self._get_com_line(*line5),
            self._get_com_line(*line6),
            self._get_com_line(*line7),
            self._get_com_line(*line8),
            self._get_com_line(*line9),
            self._get_com_line(*line10),
            '!!!BLOCK_OF_FIXED_PARAMETERS_COMES_HERE!!!',
            self._get_com_line(*line11),
            self._get_com_line(*line12),
            self._get_com_line(*line13),
        ])

        com_block = self._fix_params(mol, com_block)

        # Generate the com file containing the info for the run
        with open(f'{run_name}.com', 'w') as com:
            # name of the macromodel file
            com.write(f'{run_name}.mae\n')
            # name of the output file
            com.write(f'{run_name}-out.maegz\n')
            # details of the macromodel run
            com.write(com_block)

    def optimize(self, mol):
        """
        Optimize a molecule.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        run_name = str(uuid4().int)
        if self._output_dir is None:
            output_dir = run_name
        else:
            output_dir = self._output_dir

        mol_path = f'{run_name}.mol'

        # First write a .mol file of the molecule.
        mol.write(mol_path)
        # MacroModel requires a ``.mae`` file as input.
        self._run_structconvert(mol_path, f'{run_name}.mae')
        # Generate the ``.com`` file for the MacroModel MD run.
        self._generate_com(mol, run_name)
        # Run the optimization.
        self._run_bmin(mol, run_name)
        # Extract the lowest energy conformer into its own .mae file.
        conformer_mae = MAEExtractor(run_name).path
        mol.update_from_file(conformer_mae)

        move_generated_macromodel_files(run_name, output_dir)

    def _fix_distances(self, mol, fix_block):
        """
        Add lines fixing bond distances to ``.com`` body.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        fix_block : :class:`str`
            A string holding fix commands in the ``.com`` file.

        Returns
        -------
        :class:`str`
            A string holding fix commands in the ``.com`` file.

        """

        # Go through all the bonds in the rdkit molecule. If the bond
        # is not between bonder atoms add a fix line to the
        # ``fix_block``. If the bond does invovle two bonder atoms go
        # to the next bond. This is because a bond between 2 bonder
        # atoms was added during construction and should therefore not
        # be fixed.
        for bond in mol.bonds:
            bond_key = frozenset((bond.atom1.id, bond.atom2.id))
            if (bond_key not in self._restricted_bonds):
                continue

            # Make sure that the indices are increased by 1 in the .mae
            # file.
            atom1_id = bond.atom1.id + 1
            atom2_id = bond.atom2.id + 1
            args = ('FXDI', atom1_id, atom2_id, 0, 0, 99999, 0, 0, 0)
            fix_block += self._get_com_line(*args)
            fix_block += '\n'

        return fix_block

    def _fix_bond_angles(self, mol, fix_block):
        """
        Add lines fixing bond angles to the ``.com`` body.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        fix_block : :class:`str`
            A string holding fix commands in the ``.com`` file.

        Returns
        -------
        :class:`str`
            A string holding fix commands in the ``.com`` file.

        """

        paths = rdkit.FindAllPathsOfLengthN(
            mol=mol.to_rdkit_mol(),
            length=3,
            useBonds=False,
            useHs=True
        )
        for atom_ids in paths:
            if frozenset(atom_ids) in self._restricted_bond_angles:
                atom_ids = [i+1 for i in atom_ids]
                args = ('FXBA', *atom_ids, 99999, 0, 0, 0, 0)
                fix_block += self._get_com_line(*args)
                fix_block += '\n'

        return fix_block

    def _fix_torsional_angles(self, mol, fix_block):
        """
        Add lines fixing torsional bond angles to the ``.com`` body.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        fix_block : :class:`str`
            A string holding fix commands in the ``.com`` file.

        Returns
        -------
        :class:`str`
            A string holding fix commands in the ``.com`` file.

        """

        paths = rdkit.FindAllPathsOfLengthN(
            mol=mol.to_rdkit_mol(),
            length=4,
            useBonds=False,
            useHs=True
        )
        # Apply the fix.
        for atom_ids in paths:
            if frozenset(atom_ids) in self._restricted_torsional_angles:
                atom_ids = [i+1 for i in atom_ids]
                args = ('FXTA', *atom_ids, 99999, 361, 0, 0)
                fix_block += self._get_com_line(*args)
                fix_block += '\n'

        return fix_block



"""
Populations
===========

:class:`.Population` objects are container specialized to perform
operations on groups of ``stk`` molecules, often in parallel.

"""

import itertools as it
import os
from os.path import join
import numpy as np
import json
import psutil
from functools import wraps
import logging
import pathos

from .utilities import dedupe, dice_similarity
from .molecular import ConstructedMolecule, Molecule


logger = logging.getLogger(__name__)


class Population:
    """
    A container for  :class:`.Molecule` objects.

    :class:`Population` instances can be nested.

    In addtion to holding :class:`.Molecule` objects, the
    :class:`Population` class can be used to create large numbers of
    these instances through the class methods beginning with "init".

    :class:`.Molecule` instances held by a :class:`Population` can have
    their structures optimized in parallel through the
    :meth:`optimize` method.

    It supports all expected and necessary container operations such as
    iteration, indexing and membership checks (via the ``is in``
    operator).

    Attributes
    ----------
    direct_members : :class:`list` of :class:`.Molecule`
        Held here are direct members of the :class:`Population`.
        In other words, these are the molecules not held by any
        subpopulations. As a result, not all members of a
        :class:`Population` are stored in this attribute.

    subpopulations : :class:`list` of :class:`Population`
        A :class:`list` holding the subpopulations.

    Examples
    --------
    A :class:`Population` can be iterated through just like a
    :class:`list`

    .. code-block:: python

        import stk

        # Create a population.
        pop = stk.Population(
            stk.BuildingBlock(...),
            stk.ConstructedMolecule(...),
            stk.BuildingBlock(...),
            stk.BuildingBlock(...),

            stk.Population(
                stk.BuildingBlock(...),
                stk.ConstructedMolecule(...)
            )

            stk.ConstructedMolecule(...),
            stk.BuildingBlock(...)

        )

        for member in pop:
            do_stuff(member)

    When iterating through a :class:`Population` you will also iterate
    through nested members, that is members which are held by
    subpopulations. If you only wish to iterate through direct
    members, you can

    .. code-block:: python

        for member in pop.direct_members:
            do_stuff(member)

    You can also get access to members by using indices.
    Indices have access to all members in the population

    .. code-block:: python

        first_member = pop[0]
        second_member = pop[1]

    Indices will first access direct members of the population and then
    acess members in the subpopulations. Indices access nested members
    depth-first

    .. code-block:: python

        pop2 = stk.Population(bb1, bb2, stk.Population(bb3, bb4))
        # Get bb1.
        pop2[0]
        # Get bb2.
        pop2[1]
        # Get bb3.
        pop2[2]
        # Get bb4.
        pop2[3]

    You can get a subpopulation by taking a slice

    .. code-block:: python

        # new_pop is a new Population instance and has no nesting.
        new_pop = pop[2:4]

    You can take the length of a population to get the total number
    of members

    .. code-block:: python

        len(pop)

    Adding populations creates a new population with both of the added
    populations as subpopulations

    .. code-block:: python

        # added has no direct members and two subpopulations, pop and
        # pop2.
        added = pop + pop2

    Subtracting populations creates a new, flat population.

    .. code-block:: python

        # subbed has all objects in pop except those also found in
        # pop2.
        subbed = pop - pop2

    You can check if an object is already present in the population.

    .. code-block:: python

        bb1 = stk.BuildingBlock(...)
        bb2 = stk.BuildingBlock(...)
        pop3 = stk.Population(bb1)

        # Returns True.
        bb1 in pop3
        # Returns False.
        bb2 in pop3
        # Returns True.
        bb2 not in pop3

    """

    def __init__(self, *args):
        """
        Initialize a :class:`Population`.

        Parameters
        ----------
        *args : :class:`.Molecule`, :class:`Population`
            A population is initialized with the :class:`.Molecule` and
            :class:`Population` instances it should hold.

        Examples
        --------
        .. code-block:: python

            bb1 = stk.BuildingBlock('CCC')
            bb2 = stk.BuildingBlock('NCCNCNC')
            bb3 = stk.BuildingBlock('[Br]CCC[Br]')
            pop1 = stk.Population(bb1, bb2, bb3)

            bb4 = stk.BuildingBlock('NNCCCN')
            # pop2 has pop1 as a subpopulation and bb4 as a direct
            # member.
            pop2 = stk.Population(pop1, bb4)

        """

        self.direct_members = []
        self.subpopulations = []

        for arg in args:
            if isinstance(arg, Population):
                self.subpopulations.append(arg)
            else:
                self.direct_members.append(arg)

    @classmethod
    def init_all(
        cls,
        building_blocks,
        topology_graphs,
        num_processes=None,
        duplicates=False,
        use_cache=False
    ):
        """
        Make all possible molecules from groups of building blocks.

        Parameters
        ----------
        building_blocks : :class:`list` of :class:`.Molecule`
            A :class:`list` holding nested building blocks, for
            example

            .. code-block:: python

                bbs1 = [
                    stk.BuildingBlock(...),
                    stk.BuildingBlock(...),
                    ...
                ]
                bbs2 = [
                    stk.ConstructedMolecule(...),
                    stk.BuildingBlock(...),
                    ...,
                ]
                bbs3 = [
                    stk.BuildingBlock(...),
                    stk.BuildingBlock(...),
                    ...
                ]
                building_blocks = [bbs1, bbs2, bbs3]

            To construct a new :class:`.ConstructedMolecule`, a
            :class:`.Molecule` is picked from each of the sublists
            in `building_blocks`. The picked :class:`.Molecule`
            instances are then supplied to
            :class:`.ConstructedMolecule`

            .. code-block:: python

                # mol is a new ConstructedMolecule. bb1 is selected
                # from bbs1, bb2 is selected from bbs2 and bb3 is
                # selected from bbs3.
                mol = stk.ConstructedMolecule(
                    building_blocks=[bb1, bb2, bb3],
                    topology_graph=topology_pick
                )

            The order a :class:`.Molecule` instance is given to
            the :class:`.ConstructedMolecule` is determined by the
            sublist of `building_blocks` it was picked from. Note that
            the number of sublists in `building_blocks` is not fixed.
            It merely has to be compatible with the
            `topology_graphs`.

        topology_graphs : :class:`list` of :class:`.TopologyGraph`
            The topology graphs of `.ConstructedMolecule` being made.

        num_processes : :class:`int`, optional
            The number of parallel processes to create when
            constructing the molecules. If ``None``, creates a process
            for each core on the computer.

        duplicates : :class:`bool`, optional
            If ``False``, duplicate structures are removed from
            the population.

        use_cache : :class:`bool`, optional
            Toggles use of the molecular cache.

        Returns
        -------
        :class:`Population`
            A :class:`.Population` holding `.ConstructedMolecule`
            instances.

        Examples
        --------
        Construct all possible cage molecules from some precursors

        .. code-block:: python

            import stk

            amines = [
                stk.BuildingBlock('NCCCN', ['amine']),
                stk.BuildingBlock('NCCCCCN', ['amine']),
                stk.BuildingBlock('NCCOCCN', ['amine']),
            ]
            aldehydes = [
                stk.BuildingBlock('O=CCC(C=O)CC=O', ['aldehyde']),
                stk.BuildingBlock('O=CCC(C=O)CC=O', ['aldehyde']),
                stk.BuildingBlock('O=C(C=O)COCC=O', ['aldehyde']),
            ]
            # A total of 9 cages will be created.
            cages = stk.Population.init_all(
                building_blocks=[amines, aldehydes],
                topology_graphs=[stk.cage.FourPlusSix()]
            )

        Use the constructed cages and a new bunch of building blocks to
        create all possible cage complexes.

        .. code-block:: python

            encapsulants = [
                stk.BuildingBlock('[Br][Br]'),
                stk.BuildingBlock('[F][F]'),
            ]

            # Every combination of cage and encapsulant.
            complexes = stk.Population.init_all(
                building_blocks=[cages, encapsulants],
                topology_graphs=[stk.host_guest_complex.Complex()]
            )

        """

        bbs, topologies = [], []
        mols = it.product(*building_blocks, topology_graphs)
        for *mol_bbs, topology in mols:
            bbs.append(mol_bbs),
            topologies.append(topology)

        with pathos.pools.ProcessPool(num_processes) as pool:
            mols = pool.map(ConstructedMolecule, bbs, topologies)

        # Update the cache.
        if use_cache:
            for i, mol in enumerate(mols):
                # If the molecule did not exist already, add it to the
                # cache.
                if (
                    not ConstructedMolecule.has_cached_mol(
                        identity_key=mol.get_identity_key()
                    )
                ):
                    mol.update_cache()
                # If the molecule did exist already, use the cached
                # version.
                else:
                    mols[i] = ConstructedMolecule.get_cached_mol(
                        identity_key=mol.get_identity_key()
                    )

        p = cls(*mols)
        if not duplicates:
            p.remove_duplicates()
        return p

    def clone(self):
        """
        Return a clone.

        The clone will share the :class:`.Molecule` objects, copies of
        :class:`.Molecule` objects will not be made.

        Returns
        -------
        :class:`Population`
            The clone.

        Examples
        --------
        .. code-block:: python

            import stk

            # Make an intial population.
            pop = stk.Population(stk.BuildingBlock('NCCN'))
            # Make a clone.
            clone = pop.clone()

        """

        clone = self.__class__(*self.direct_members)
        clone.subpopulations = [
            pop.clone() for pop in self.subpopulations
        ]
        return clone

    @classmethod
    def init_diverse(
        cls,
        building_blocks,
        topology_graphs,
        size,
        random_seed=None,
        use_cache=False
    ):
        """
        Construct a chemically diverse :class:`.Population`.

        All constructed molecules are held in :attr:`direct_members`.

        In order to construct a :class:`.ConstructedMolecule`, a random
        :class:`.Molecule` is selected from each sublist in
        `building_blocks`. Once the first construction is complete,
        the next :class:`.Molecule` selected from each sublist is the
        one with the most different Morgan fingerprint to the prior
        one. The third construction uses randomly selected
        :class:`.Molecule` objects again and so on. This is done until
        `size` :class:`.ConstructedMolecule` instances have been
        constructed.

        Parameters
        ----------
        building_blocks : :class:`list` of :class:`.Molecule`
            A :class:`list` holding nested building blocks, for
            example

            .. code-block:: python

                bbs1 = [
                    stk.BuildingBlock(...),
                    stk.BuildingBlock(...),
                    ...
                ]
                bbs2 = [
                    stk.ConstructedMolecule(...),
                    stk.BuildingBlock(...),
                    ...
                ]
                bbs3 = [
                    stk.BuildingBlock(...),
                    stk.BuildingBlock(...),
                    ...
                ]
                building_blocks = [bbs1, bbs2, bbs3]

            To construct a new :class:`.ConstructedMolecule`, a
            :class:`.Molecule` is picked from each of the sublists
            in `building_blocks`. The picked :class:`.Molecule`
            instances are then supplied to the
            :class:`.ConstructedMolecule`

            .. code-block:: python

                # mol is a new ConstructedMolecule. bb1 is selected
                # from bbs1, bb2 is selected from bbs2 and bb3 is
                # selected from bbs3.
                mol = stk.ConstructedMolecule(
                    building_blocks=[bb1, bb2, bb3],
                    topology_graph=topology_pick
                )

            The order a :class:`.Molecule` instance is given to
            :class:`.ConstructedMolecule` is determined by the
            sublist of `building_blocks` it was picked from. Note that
            the number of sublists in `building_blocks` is not fixed.
            It merely has to be compatible with the `topology_graphs`.

        topology_graphs : :class:`iterable` of :class:`.TopologyGraph`
            An iterable holding topology grpahs which should be
            randomly selected for the construction of a
            :class:`.ConstructedMolecule`.

        size : :class:`int`
            The desired size of the :class:`.Population`.

        random_seed : :class:`int`, optional
            Seed for the random number generator to get replicable
            results.

        use_cache : :class:`bool`, optional
            Toggles use of the molecular cache.

        Returns
        -------
        :class:`.Population`
            A population filled with the constructed molecules.

        Examples
        --------
        Construct a diverse :class:`.Population` of cage
        molecules from some precursors

        .. code-block:: python

            import stk

            amines = [
                stk.BuildingBlock('NCCCN', ['amine']),
                stk.BuildingBlock('NCCCCCN', ['amine']),
                stk.BuildingBlock('NCCOCCN', ['amine']),
            ]
            aldehydes = [
                stk.BuildingBlock('O=CCC(C=O)CC=O', ['aldehyde']),
                stk.BuildingBlock('O=CCC(C=O)CC=O', ['aldehyde']),
                stk.BuildingBlock('O=C(C=O)COCC=O', ['aldehyde']),
            ]
            # A total of 4 cages will be created.
            cages = stk.Population.init_diverse(
                building_blocks=[amines, aldehydes],
                topology_graphs=[stk.cage.FourPlusSix()],
                size=4
            )

        Use the constructed cages and a new bunch of building blocks to
        create some diverse cage complexes.

        .. code-block:: python

            encapsulants = [
                stk.BuildingBlock('[Br][Br]'),
                stk.BuildingBlock('[F][F]'),
            ]

            # 4 combinations of cage and encapsulant.
            complexes = stk.Population.init_diverse(
                building_blocks=[cages, encapsulants],
                topology_graphs=[stk.host_guest_complex.Complex()],
                size=4
            )

        """

        pop = cls()

        # Shuffle the sublists.
        generator = np.random.RandomState(random_seed)
        for db in building_blocks:
            generator.shuffle(db)

        # Go through every possible constructed molecule.
        for *bbs, top in it.product(*building_blocks, topology_graphs):

            # Generate the random constructed molecule.
            mol = ConstructedMolecule(
                building_blocks=bbs,
                topology_graph=top,
                use_cache=use_cache
            )
            if mol not in pop:
                pop.direct_members.append(mol)

            if len(pop) == size:
                break

            # Get the most different BuildingBlock to the previously
            # selected one, per sublist.
            diff_bbs = [
                min(db, key=lambda mol: dice_similarity(bb, mol))
                for bb, db in zip(bbs, building_blocks)

            ]
            mol = ConstructedMolecule(
                building_blocks=diff_bbs,
                topology_graph=top,
                use_cache=use_cache
            )
            if mol not in pop:
                pop.direct_members.append(mol)

            if len(pop) == size:
                break

        assert len(pop) == size
        return pop

    @classmethod
    def init_random(
        cls,
        building_blocks,
        topology_graphs,
        size,
        random_seed=None,
        use_cache=False
    ):
        """
        Construct molecules for a random :class:`.Population`.

        All molecules are held in :attr:`direct_members`.

        From the supplied building blocks a random :class:`.Molecule`
        is selected from each sublist to form a
        :class:`.ConstructedMolecule`. This is done until `size`
        :class:`.ConstructedMolecule` objects have been constructed.

        Parameters
        ----------
        building_blocks : :class:`list` of :class:`.Molecule`
            A :class:`list` holding nested building blocks, for
            example

            .. code-block:: python

                bbs1 = [
                    stk.BuildingBlock(...),
                    sk.BuildingBlock(...),
                    ...
                ]
                bbs2 = [
                    stk.ConstructedMolecule(...),
                    stk.BuildingBlock(...),
                    ...
                ]
                bbs3 = [
                    stk.BuildingBlock(...),
                    stk.BuildingBlock(...),
                    ...
                ]
                building_blocks = [bbs1, bbs2, bbs3]

            To construct a new :class:`.ConstructedMolecule`, a
            :class:`.Molecule` is picked from each of the sublists
            in `building_blocks`. The picked :class:`.Molecule`
            instances are then supplied to
            :class:`.ConstructedMolecule`

            .. code-block:: python

                # mol is a new ConstructedMolecule. bb1 is selected
                # from bbs1, bb2 is selected from bbs2 and bb3 is
                # selected from bbs3.
                mol = stk.ConstructedMolecule(
                    building_blocks=[bb1, bb2, bb3],
                    topology_graph=topology_pick
                )

            The order a :class:`.Molecule` instance is given to
            the :class:`.ConstructedMolecule` is determined by the
            sublist of `building_blocks` it was picked from. Note that
            the number of sublists in `building_blocks` is not fixed.
            It merely has to be compatible with the `topology_graphs`.

        topology_graphs : :class:`iterable` of :class:`.TopologyGraph`
            An :class:`iterable` holding topology graphs which should
            be randomly selected during initialization of
            :class:`.ConstructedMolecule`.

        size : :class:`int`
            The size of the population to be initialized.

        random_seed : :class:`int`, optional
            Seed for the random number generator to get replicable
            results.

        use_cache : :class:`bool`, optional
            Toggles use of the molecular cache.

        Returns
        -------
        :class:`.Population`
            A population filled with random
            :class:`.ConstructedMolecule` instances.

        Examples
        --------
        Construct 5 random cage molecules from some
        precursors

        .. code-block:: python

            import stk

            amines = [
                stk.BuildingBlock('NCCCN', ['amine']),
                stk.BuildingBlock('NCCCCCN', ['amine']),
                stk.BuildingBlock('NCCOCCN', ['amine']),
            ]
            aldehydes = [
                stk.BuildingBlock('O=CCC(C=O)CC=O', ['aldehyde']),
                stk.BuildingBlock('O=CCC(C=O)CC=O', ['aldehyde']),
                stk.BuildingBlock('O=C(C=O)COCC=O', ['aldehyde']),
            ]
            # A total of 5 cages will be created.
            cages = stk.Population.init_random(
                building_blocks=[amines, aldehydes],
                topology_graphs=[stk.cage.FourPlusSix()],
                size=5
            )

        Use the constructed cages and a new bunch of building blocks to
        create some random cage complexes.

        .. code-block:: python

            encapsulants = [
                stk.BuildingBlock('[Br][Br]'),
                stk.BuildingBlock('[F][F]'),
            ]

            # Random combinations of cage and encapsulant.
            complexes = stk.Population.init_random(
                building_blocks=[cages, encapsulants],
                topology_graphs=[stk.host_guest_complex.Complex()],
                size=5
            )

        """

        pop = cls()
        generator = np.random.RandomState(random_seed)

        # Shuffle the sublists.
        for db in building_blocks:
            generator.shuffle(db)

        # Go through every possible constructed molecule.
        for *bbs, top in it.product(*building_blocks, topology_graphs):
            # Generate the random constructed molecule.
            mol = ConstructedMolecule(
                building_blocks=bbs,
                topology_graph=top,
                use_cache=use_cache
            )
            if mol not in pop:
                pop.direct_members.append(mol)

            if len(pop) == size:
                break

        assert len(pop) == size
        return pop

    @classmethod
    def init_from_list(cls, pop_list, use_cache=False):
        """
        Initialize a population from a :class:`list` representation.

        Parameters
        ----------
        pop_list : :class:`list`
            A :class:`list` which represents a :class:`Population`.
            Like the ones created by :meth:`to_list`. For example in,

            .. code-block:: python

                pop_list = [{...}, [{...}], [{...}, {...}], {...}]

            ``pop_list`` represents the :class:`Population`, sublists
            represent its subpopulations and the :class:`dict`
            ``{...}`` represents the members.

        use_cache : :class:`bool`, optional
            Toggles use of the molecular cache.

        Returns
        -------
        :class:`Population`
            The population represented by `pop_list`.

        """

        pop = cls()
        for item in pop_list:
            if isinstance(item, list):
                sp = cls.init_from_list(item, use_cache=use_cache)
                pop.subpopulations.append(sp)
            else:
                pop.direct_members.append(
                    Molecule.init_from_dict(item, use_cache=use_cache)
                )
        return pop

    def add_members(self, molecules, duplicate_key=None):
        """
        Add :class:`.Molecule` instances to the :class:`.Population`.

        The added :class:`.Molecule` instances are added as direct
        members of the population, they are not placed into any
        subpopulations.

        Parameters
        ----------
        molecules : :class:`iterable` of :class:`.Molecule`
            The molecules to be added as direct members.

        duplicate_key : :class:`callable`, optional
            If not ``None``, ``duplicate_key(mol)`` is evalued on each
            molecule in `members`. If a molecule with the same
            `duplicate_key` is already present in the population, the
            molecule is not added.

        Returns
        -------
        None : :class:`NoneType`

        """

        if duplicate_key is None:
            self.direct_members.extend(molecules)
        else:
            keys = {duplicate_key(mol) for mol in self}
            for mol in molecules:
                key = duplicate_key(mol)
                if key not in keys:
                    keys.add(key)
                    self.direct_members.append(mol)

    def add_subpopulation(self, population):
        """
        Add a clone of `population` to :attr:`subpopulations`.

        Only a clone of the `population` container is made. The
        molecules it holds are not copies.

        Parameters
        ----------
        population : :class:`Population`
            The population to be added as a subpopulation.

        Returns
        -------
        None : :class:`NoneType`

        """

        self.subpopulations.append(population.clone())

    def _get_all_members(self):
        """
        Yield both direct and nested members of the population.

        Yields
        ------
        :class:`.Molecule`
            The next :class:`.Molecule` instance held within the
            :class:`.Population`.

        """

        # Go through `direct_members` attribute and yield ``Molecule``
        # instances held within one by one.
        for ind in self.direct_members:
            yield ind

        # Go thorugh `populations` attribute and for each
        # ``Population`` instance within, yield ``Molecule``
        # instances from its `all_members` generator.
        for pop in self.subpopulations:
            yield from pop._get_all_members()

    def set_mol_ids(self, n, overwrite=False):
        """
        Give each member of the population an id starting from `n`.

        This method adds an :attr:`id` attribute to each
        :class:`.Molecule` instance held by the population.

        Parameters
        ----------
        n : :class:`int`
            A number. Members of this :class:`Population` are given a
            unique number as an id, starting from `n` and
            incremented by one between members.

        overwrite : :class:`bool`, optional
            If ``True``, existing ids are replaced.

        Returns
        -------
        :class:`int`
            The value of the last id assigned, plus 1.

        """

        for mem in self:
            if not hasattr(mem, 'id') or overwrite:
                mem.id = n
                n += 1
        return n

    def dump(
        self,
        path,
        include_attrs=None,
        ignore_missing_attrs=False
    ):
        """
        Dump the :class:`.Population` to a file.

        Parameters
        ----------
        path : :class:`str`
            The full path of the file to which the :class:`Population`
            should be dumped.

        include_attrs : :class:`list` of :class:`str`, optional
            The names of attributes of the molecules to be added to
            the JSON. Each attribute is saved as a string using
            :func:`repr`.

        ignore_missing_attrs : :class:`bool`, optional
            If ``False`` and an attribute in `include_attrs` is not
            held by a :class:`.Molecule`, an error will be raised.

        Returns
        -------
        None : :class:`NoneType`

        """

        with open(path, 'w') as f:
            content = self.to_list(include_attrs, ignore_missing_attrs)
            json.dump(content, f, indent=4)

    @classmethod
    def load(cls, path, use_cache=False):
        """
        Initialize a :class:`Population` from one dumped to a file.

        Parameters
        ----------
        path : :class:`str`
            The full path of the file holding the dumped population.

        use_cache : :class:`bool`, optional
            Toggles use of the moleular cache.

        Returns
        -------
        :class:`Population`
            The population stored in the dump file.

        """

        with open(path, 'r') as f:
            pop_list = json.load(f)

        return cls.init_from_list(pop_list, use_cache)

    def _optimize_parallel(self, optimizer, num_processes):
        opt_fn = _Guard(optimizer, optimizer.optimize)

        # Apply the function to every member of the population, in
        # parallel.
        with pathos.pools.ProcessPool(num_processes) as pool:
            optimized = pool.map(opt_fn, self)

        # If anything failed, raise an error.
        for result in optimized:
            if isinstance(result, Exception):
                raise result

        # Update the structures in the population.
        sorted_opt = sorted(optimized, key=lambda m: repr(m))
        sorted_pop = sorted(self, key=lambda m: repr(m))
        for old, new in zip(sorted_pop, sorted_opt):
            assert old.get_identity_key() == new.get_identity_key()
            old.__dict__ = dict(vars(new))
            if optimizer.is_caching():
                optimizer.add_to_cache(old)

    def _optimize_serial(self, optimizer):
        for member in self:
            optimizer.optimize(member)

    def optimize(self, optimizer, num_processes=None):
        """
        Optimize the structures of molecules in the population.

        The molecules are optimized serially or in parallel depending
        if `num_processes` is ``1`` or more. The serial version may be
        faster in cases where all molecules have already been
        optimized and the `optimizer` will skip them.
        In this case creating a parallel process pool creates
        unncessary overhead.

        Parameters
        ----------
        optimizer : :class:`.Optimizer`
            The optimizer used to carry out the optimizations.

        num_processes : :class:`int`, optional
            The number of parallel processes to create. Optimization
            will run serially if ``1``. If ``None``, creates a
            process for each core on the computer.

        Returns
        -------
        None : :class:`NoneType`

        """

        if num_processes is None:
            num_processes = psutil.cpu_count()

        if num_processes == 1:
            self._optimize_serial(optimizer)
        else:
            self._optimize_parallel(optimizer, num_processes)

    def remove_duplicates(self, across_subpopulations=True, key=id):
        """
        Remove duplicates from the population.

        The question of which molecule is preserved when duplicates are
        removed is difficult to answer. The iteration through a
        population is depth-first, so a rule such as "the molecule in
        the topmost population is preserved" is not the case here.
        Rather, the first molecule found is preserved.

        However, this question is only relevant if duplicates in
        different subpopulations are being removed. In this case it is
        assumed that it is more important to have a single instance
        than to worry about which subpopulation it is in.

        If the duplicates are being removed from within subpopulations,
        each subpopulation will end up with a single instance of all
        molecules held before. There is no "choice".

        Parameters
        ----------
        across_subpopulations : :class:`bool`, optional
            When ``False`` duplicates are only removed from within a
            given subpopulation. If ``True``, all duplicates are
            removed, regardless of which subpopulation they are in.

        key : :class:`callable`, optional
            Two molecules are considered the same if the values
            returned by ``key(molecule)`` are the same.

        Returns
        -------
        None : :class:`NoneType`

        """

        return self._remove_duplicates(
            across_subpopulations=across_subpopulations,
            key=key,
            seen=set()
        )

    def _remove_duplicates(self, across_subpopulations, key, seen):
        if not across_subpopulations:
            seen = set()

        self.direct_members = list(
            dedupe(self.direct_members, key, seen)
        )
        for subpop in self.subpopulations:
            subpop._remove_duplicates(across_subpopulations, key, seen)

    def remove_members(self, key):
        """
        Remove all members where ``key(member)`` is ``True``.

        Parameters
        ----------
        key : :class:`callable`
            A callable which takes 1 argument. Each member of the
            population is passed as the argument to `key` in turn. If
            the result is ``True`` then the member is removed from the
            population.

        Returns
        -------
        None : :class:`NoneType`

        """

        self.direct_members = [
            ind for ind in self.direct_members if not key(ind)
        ]
        for subpop in self.subpopulations:
            subpop.remove_members(key)

    def to_list(self, include_attrs=None, ignore_missing_attrs=False):
        """
        Convert the population to a :class:`list` representation.

        Parameters
        ----------
        include_attrs : :class:`list` of :class:`str`, optional
            The names of attributes to be added to the molecular
            representations. Each attribute is saved as a string using
            :func:`repr`.

        ignore_missing_attrs : :class:`bool`, optional
            If ``False`` and an attribute in `include_attrs` is not
            held by a :class:`.Molecule`, an error will be raised.

        Returns
        -------
        :class:`list`
            A :class:`list` representation of the :class:`Population`.

        """

        if include_attrs is None:
            include_attrs = []

        pop = [
            m.to_dict(include_attrs, ignore_missing_attrs)
            for m in self.direct_members
        ]
        for sp in self.subpopulations:
            pop.append(sp.to_list(include_attrs, ignore_missing_attrs))
        return pop

    def write(self, path):
        """
        Write the ``.mol`` files of members to a directory.

        Parameters
        ----------
        path : :class:`str`
            The full path of the directory into which the ``.mol`` file
            is written.

        Returns
        -------
        None : :class:`NoneType`

        """

        # If the directory does not exist, create it.
        if not os.path.exists(path):
            os.mkdir(path)

        for i, member in enumerate(self):
            member.write(join(path, f'{i}.mol'))

    def __iter__(self):
        """
        Iterate through members of the :class:`Population`.

        When :class:`Population` instances are iterated through, they
        yield both direct and nested members.

        Yields
        ------
        :class:`.Molecule`
            The next molecule in the population.

        """

        yield from self._get_all_members()

    def __getitem__(self, key):
        """
        Get a member or members by index.

        Molecules held by the :class:`Population` instance can be
        accessed by index. Slices are also supported and they return
        a new :class:`Population` instance holding the members with the
        requested indices. Using slices will return a flat
        :class:`Population` instance, meaing no nesting is preserved.

        Indexing accesses both direct and nested members of the
        :class:`Population`.

        Parameters
        ----------
        key : :class:`int`, :class:`slice`
            An :class:`int` or :class:`slice` can be used depending on
            if a single members needs to be returned or a collection of
            them.

        Returns
        -------
        :class:`.Molecule`
            If the supplied `key` is an :class:`int`. Returns the
            :class:`.Molecule` instance at the corresponding index.

        :class:`Population`
            If the supplied `key` is a :class:`slice`. The returned
            :class:`Population` instance holds members at the given
            indices.

        Raises
        ------
        :class:`IndexError`
            If the supplied `key` is out of range.

        :class:`TypeError`
            If the supplied `key` is not an :class:`int` or
            :class:`slice`.

        """

        if isinstance(key, int):
            for i, member in enumerate(self):
                if i == key:
                    return member
            raise IndexError('Population index out of range.')

        if isinstance(key, slice):
            mols = it.islice(self, key.start, key.stop, key.step)
            return self.__class__(*mols)

        raise TypeError(
            'Index must be an integer or slice, not '
            f'{key.__class__.__name__}.'
        )

    def __len__(self):
        """
        Return the total number of members in the population.

        Returns
        -------
        :class:`int`
            The number of members held by the population, including
            those held within its subpopulations.

        """

        size = len(self.direct_members)
        stack = list(self.subpopulations)
        while stack:
            subpop = stack.pop()
            stack.extend(subpop.subpopulations)
            size += len(subpop.direct_members)
        return size

    def __sub__(self, other):
        """
        Remove members of `other` from the population.

        Subtracting one population from another,

        .. code-block:: python

            pop3 = pop1 - pop2

        returns a new population, ``pop3``. The returned population
        contains all molecules in ``pop1`` except those also found in
        ``pop2``. This refers to all molecules, including those held
        within any subpopulations. The returned population is flat.
        This means any nesting in ``pop1`` is not preserved.

        Parameters
        ----------
        other : :class:`Population`
            A collection of :class:`.Molecule` instances to be removed
            from the population, if held by it.

        Returns
        -------
        :class:`Population`
            A flat population of :class:`.Molecule` instances which
            are in the population but not in `other`.

        """

        new_pop = self.__class__()
        new_pop.add_members(
            molecules=(mol for mol in self if mol not in other)
        )
        return new_pop

    def __add__(self, other):
        """
        Add two populations.

        Parameters
        ----------
        other : :class:`Population`
            A population to be joined with.

        Returns
        -------
        :class:`Population`
            A new :class:`Population` instance which holds two
            subpopulations and no direct members. The two
            subpopulations are the two :class:`Population` instances on
            which the ``+`` operator was applied.

        """

        return self.__class__(self, other)

    def __contains__(self, item):
        return any(mol is item for mol in self)

    def __str__(self):
        name = f'Population {id(self)}\n'
        dash = '-'
        members = 'Members'
        subpopulations = 'Subpopulations'

        if self.direct_members:
            direct_members = '\n'.join(
                f'\t{mol}' for mol in self.direct_members
            )
        else:
            direct_members = '\tNone'

        if self.subpopulations:
            subpops = ', '.join(
                str(id(sp)) for sp in self.subpopulations
            )
        else:
            subpops = 'None'

        subpop_strs = '\n'.join(str(sp) for sp in self.subpopulations)

        return (
            f'{name}\n{dash*len(name)}\n'
            f'\t{members}\n\t{dash*len(members)}\n'
            f'{direct_members}\n\n'
            f'\t{subpopulations}\n\t{dash*len(subpopulations)}\n'
            f'{subpops}\n'
            f'{subpop_strs}'
        )

    def __repr__(self):
        return str(self)


class _Guard:
    """
    A decorator for optimization functions.

    This decorator should be applied to all functions which are to
    be used with :mod:`pathos`. It prevents functions from
    raising if they fail, which prevents the :mod:`pathos` pool
    from hanging.

    Attributes
    ----------
    _calc_name : :class:`str`
        The name of the :class:`.Optimizer` class being used.

    """

    def __init__(self, calculator, fn):
        self._calc_name = calculator.__class__.__name__
        wraps(fn)(self)

    def __call__(self, mol):
        """
        Decorates and calls the function.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        :class:`.Molecule`
            The optimized molecule.

        """

        fn = self.__wrapped__.__name__
        cls = self._calc_name
        try:
            logger.info(f'Running "{cls}.{fn}()" on "{mol}"')
            self.__wrapped__(mol)
            return mol

        except Exception as ex:
            errormsg = (
                f'"{cls}.{fn}()" failed on molecule "{mol}"'
            )
            logger.error(errormsg, exc_info=True)
            return ex



"""
Defines cage topologies from 2 and 3 functionalized building blocks.

"""

import numpy as np

from .base import Cage, _CageVertexData
from ..topology_graph import EdgeData


class TwoPlusThree(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    vertex_data = (
        _CageVertexData(0, 0, 1),
        _CageVertexData(0, 0, -1),

        _CageVertexData(-1, -0.5*np.sqrt(3), 0, False),
        _CageVertexData(1, -0.5*np.sqrt(3), 0, False),
        _CageVertexData(0, 0.5*np.sqrt(3), 0, False)

    )

    edge_data = (
        EdgeData(vertex_data[0], vertex_data[2]),
        EdgeData(vertex_data[0], vertex_data[3]),
        EdgeData(vertex_data[0], vertex_data[4]),
        EdgeData(vertex_data[1], vertex_data[2]),
        EdgeData(vertex_data[1], vertex_data[3]),
        EdgeData(vertex_data[1], vertex_data[4])
    )

    num_windows = 3
    num_window_types = 1


class FourPlusSix(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    # Vertices of a tetrahdron so that origin is at the origin. Source:
    # http://tinyurl.com/lc262h8.
    _v0, _v1, _v2, _v3 = _vertex_data = (
        _CageVertexData(0, 0, np.sqrt(6)/2),
        _CageVertexData(-1, -np.sqrt(3)/3, -np.sqrt(6)/6),
        _CageVertexData(1, -np.sqrt(3)/3, -np.sqrt(6)/6),
        _CageVertexData(0, 2*np.sqrt(3)/3, -np.sqrt(6)/6)
    )

    vertex_data = (
        *_vertex_data,
        _CageVertexData.init_at_center(_v0, _v1),
        _CageVertexData.init_at_center(_v0, _v2),
        _CageVertexData.init_at_center(_v0, _v3),
        _CageVertexData.init_at_center(_v1, _v2),
        _CageVertexData.init_at_center(_v1, _v3),
        _CageVertexData.init_at_center(_v2, _v3)

    )

    edge_data = (
        EdgeData(vertex_data[4], vertex_data[0]),
        EdgeData(vertex_data[4], vertex_data[1]),
        EdgeData(vertex_data[5], vertex_data[0]),
        EdgeData(vertex_data[5], vertex_data[2]),
        EdgeData(vertex_data[6], vertex_data[0]),
        EdgeData(vertex_data[6], vertex_data[3]),
        EdgeData(vertex_data[7], vertex_data[1]),
        EdgeData(vertex_data[7], vertex_data[2]),
        EdgeData(vertex_data[8], vertex_data[1]),
        EdgeData(vertex_data[8], vertex_data[3]),
        EdgeData(vertex_data[9], vertex_data[2]),
        EdgeData(vertex_data[9], vertex_data[3])
    )

    num_windows = 4
    num_window_types = 1


class FourPlusSix2(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _vertex_data = (
        _CageVertexData(1, 0, 1),
        _CageVertexData(-1, 0, 1),
        _CageVertexData(1, 0, -1),
        _CageVertexData(-1, 0, -1),

        _CageVertexData(0, -1, 1, False),
        _CageVertexData(0, 1, 1, False),
        _CageVertexData(0, -1, -1, False),
        _CageVertexData(0, 1, -1, False)
    )

    vertex_data = (
        *_vertex_data,
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[2]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[3]
        )
    )

    edge_data = (
        EdgeData(vertex_data[4], vertex_data[0]),
        EdgeData(vertex_data[4], vertex_data[1]),
        EdgeData(vertex_data[5], vertex_data[0]),
        EdgeData(vertex_data[5], vertex_data[1]),
        EdgeData(vertex_data[6], vertex_data[2]),
        EdgeData(vertex_data[6], vertex_data[3]),
        EdgeData(vertex_data[7], vertex_data[2]),
        EdgeData(vertex_data[7], vertex_data[3]),
        EdgeData(vertex_data[8], vertex_data[0]),
        EdgeData(vertex_data[8], vertex_data[2]),
        EdgeData(vertex_data[9], vertex_data[1]),
        EdgeData(vertex_data[9], vertex_data[3])
    )

    num_windows = 4
    num_window_types = 2


class SixPlusNine(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    # source: http://eusebeia.dyndns.org/4d/prism3
    _vertex_data = (
        _CageVertexData(-1, -1/np.sqrt(3), -1),
        _CageVertexData(-1, -1/np.sqrt(3), 1),
        _CageVertexData(1, -1/np.sqrt(3), -1),
        _CageVertexData(1, -1/np.sqrt(3), 1),
        _CageVertexData(0, 2/np.sqrt(3), -1),
        _CageVertexData(0, 2/np.sqrt(3), 1)
    )
    vertex_data = (
        *_vertex_data,
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[1]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[2]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[2], _vertex_data[3]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[3]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[4]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[2], _vertex_data[4]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[5]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[3], _vertex_data[5]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[5]
        )
    )

    edge_data = (
        EdgeData(vertex_data[6], vertex_data[0]),
        EdgeData(vertex_data[6], vertex_data[1]),
        EdgeData(vertex_data[7], vertex_data[0]),
        EdgeData(vertex_data[7], vertex_data[2]),
        EdgeData(vertex_data[8], vertex_data[2]),
        EdgeData(vertex_data[8], vertex_data[3]),
        EdgeData(vertex_data[9], vertex_data[1]),
        EdgeData(vertex_data[9], vertex_data[3]),
        EdgeData(vertex_data[10], vertex_data[0]),
        EdgeData(vertex_data[10], vertex_data[4]),
        EdgeData(vertex_data[11], vertex_data[2]),
        EdgeData(vertex_data[11], vertex_data[4]),
        EdgeData(vertex_data[12], vertex_data[1]),
        EdgeData(vertex_data[12], vertex_data[5]),
        EdgeData(vertex_data[13], vertex_data[3]),
        EdgeData(vertex_data[13], vertex_data[5]),
        EdgeData(vertex_data[14], vertex_data[4]),
        EdgeData(vertex_data[14], vertex_data[5])
    )

    num_windows = 5
    num_window_types = 1


class EightPlusTwelve(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _vertex_data = (
        _CageVertexData(-1, 1, -1),
        _CageVertexData(-1, -1, -1),
        _CageVertexData(1, 1, -1),
        _CageVertexData(1, -1, -1),

        _CageVertexData(-1, 1, 1),
        _CageVertexData(-1, -1, 1),
        _CageVertexData(1, 1, 1),
        _CageVertexData(1, -1, 1)
    )

    vertex_data = (
        *_vertex_data,

        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[2]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[1]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[3]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[2], _vertex_data[3]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[6]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[5]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[5], _vertex_data[7]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[6], _vertex_data[7]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[4]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[5]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[2], _vertex_data[6]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[3], _vertex_data[7]
        )
    )

    edge_data = (
        EdgeData(vertex_data[8], vertex_data[0]),
        EdgeData(vertex_data[8], vertex_data[2]),

        EdgeData(vertex_data[9], vertex_data[0]),
        EdgeData(vertex_data[9], vertex_data[1]),

        EdgeData(vertex_data[10], vertex_data[1]),
        EdgeData(vertex_data[10], vertex_data[3]),

        EdgeData(vertex_data[11], vertex_data[2]),
        EdgeData(vertex_data[11], vertex_data[3]),

        EdgeData(vertex_data[12], vertex_data[4]),
        EdgeData(vertex_data[12], vertex_data[6]),

        EdgeData(vertex_data[13], vertex_data[4]),
        EdgeData(vertex_data[13], vertex_data[5]),

        EdgeData(vertex_data[14], vertex_data[5]),
        EdgeData(vertex_data[14], vertex_data[7]),

        EdgeData(vertex_data[15], vertex_data[6]),
        EdgeData(vertex_data[15], vertex_data[7]),

        EdgeData(vertex_data[16], vertex_data[0]),
        EdgeData(vertex_data[16], vertex_data[4]),

        EdgeData(vertex_data[17], vertex_data[1]),
        EdgeData(vertex_data[17], vertex_data[5]),

        EdgeData(vertex_data[18], vertex_data[2]),
        EdgeData(vertex_data[18], vertex_data[6]),

        EdgeData(vertex_data[19], vertex_data[3]),
        EdgeData(vertex_data[19], vertex_data[7])

    )

    num_windows = 6
    num_window_types = 1


class TwentyPlusThirty(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    # Source: http://tinyurl.com/h2dl949
    _phi = (1 + np.sqrt(5))/2
    _x = 1.5
    _vertex_data = (
        _CageVertexData(_x*_phi, 0.0, _x/_phi),
        _CageVertexData(_x*-_phi, 0.0, _x/_phi),
        _CageVertexData(_x*-_phi, 0.0, _x/-_phi),
        _CageVertexData(_x*_phi, 0.0, _x/-_phi),

        _CageVertexData(_x/_phi, _x*_phi, 0.0),
        _CageVertexData(_x/_phi, _x*-_phi, 0.0),
        _CageVertexData(_x/-_phi, _x*-_phi, 0.0),
        _CageVertexData(_x/-_phi, _x*_phi, 0.0),
        _CageVertexData(0.0, _x/_phi, _x*_phi),
        _CageVertexData(0.0, _x/_phi, _x*-_phi),
        _CageVertexData(0.0, _x/-_phi, _x*-_phi),
        _CageVertexData(0.0, _x/-_phi, _x*_phi),

        _CageVertexData(_x, _x, _x),
        _CageVertexData(_x, -_x, _x),
        _CageVertexData(-_x, -_x, _x),
        _CageVertexData(-_x, _x, _x),
        _CageVertexData(-_x, _x, -_x),
        _CageVertexData(_x, _x, -_x),
        _CageVertexData(_x, -_x, -_x),
        _CageVertexData(-_x, -_x, -_x)
    )

    vertex_data = (
        *_vertex_data,
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[13]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[12]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[3]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[14]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[15]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[2]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[2], _vertex_data[19]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[2], _vertex_data[16]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[3], _vertex_data[18]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[3], _vertex_data[17]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[12]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[7]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[17]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[5], _vertex_data[6]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[5], _vertex_data[18]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[5], _vertex_data[13]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[6], _vertex_data[14]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[6], _vertex_data[19]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[7], _vertex_data[15]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[7], _vertex_data[16]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[8], _vertex_data[11]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[8], _vertex_data[12]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[8], _vertex_data[15]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[9], _vertex_data[10]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[9], _vertex_data[17]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[9], _vertex_data[16]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[10], _vertex_data[18]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[10], _vertex_data[19]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[11], _vertex_data[14]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[11], _vertex_data[13])
    )

    edge_data = (
        EdgeData(vertex_data[20], vertex_data[0]),
        EdgeData(vertex_data[20], vertex_data[13]),
        EdgeData(vertex_data[21], vertex_data[0]),
        EdgeData(vertex_data[21], vertex_data[12]),
        EdgeData(vertex_data[22], vertex_data[0]),
        EdgeData(vertex_data[22], vertex_data[3]),

        EdgeData(vertex_data[23], vertex_data[1]),
        EdgeData(vertex_data[23], vertex_data[14]),
        EdgeData(vertex_data[24], vertex_data[1]),
        EdgeData(vertex_data[24], vertex_data[15]),
        EdgeData(vertex_data[25], vertex_data[1]),
        EdgeData(vertex_data[25], vertex_data[2]),

        EdgeData(vertex_data[26], vertex_data[2]),
        EdgeData(vertex_data[26], vertex_data[19]),
        EdgeData(vertex_data[27], vertex_data[2]),
        EdgeData(vertex_data[27], vertex_data[16]),

        EdgeData(vertex_data[28], vertex_data[3]),
        EdgeData(vertex_data[28], vertex_data[18]),
        EdgeData(vertex_data[29], vertex_data[3]),
        EdgeData(vertex_data[29], vertex_data[17]),

        EdgeData(vertex_data[30], vertex_data[4]),
        EdgeData(vertex_data[30], vertex_data[12]),
        EdgeData(vertex_data[31], vertex_data[4]),
        EdgeData(vertex_data[31], vertex_data[7]),
        EdgeData(vertex_data[32], vertex_data[4]),
        EdgeData(vertex_data[32], vertex_data[17]),

        EdgeData(vertex_data[33], vertex_data[5]),
        EdgeData(vertex_data[33], vertex_data[6]),
        EdgeData(vertex_data[34], vertex_data[5]),
        EdgeData(vertex_data[34], vertex_data[18]),
        EdgeData(vertex_data[35], vertex_data[5]),
        EdgeData(vertex_data[35], vertex_data[13]),

        EdgeData(vertex_data[36], vertex_data[6]),
        EdgeData(vertex_data[36], vertex_data[14]),
        EdgeData(vertex_data[37], vertex_data[6]),
        EdgeData(vertex_data[37], vertex_data[19]),

        EdgeData(vertex_data[38], vertex_data[7]),
        EdgeData(vertex_data[38], vertex_data[15]),
        EdgeData(vertex_data[39], vertex_data[7]),
        EdgeData(vertex_data[39], vertex_data[16]),

        EdgeData(vertex_data[40], vertex_data[8]),
        EdgeData(vertex_data[40], vertex_data[11]),
        EdgeData(vertex_data[41], vertex_data[8]),
        EdgeData(vertex_data[41], vertex_data[12]),
        EdgeData(vertex_data[42], vertex_data[8]),
        EdgeData(vertex_data[42], vertex_data[15]),

        EdgeData(vertex_data[43], vertex_data[9]),
        EdgeData(vertex_data[43], vertex_data[10]),
        EdgeData(vertex_data[44], vertex_data[9]),
        EdgeData(vertex_data[44], vertex_data[17]),
        EdgeData(vertex_data[45], vertex_data[9]),
        EdgeData(vertex_data[45], vertex_data[16]),

        EdgeData(vertex_data[46], vertex_data[10]),
        EdgeData(vertex_data[46], vertex_data[18]),
        EdgeData(vertex_data[47], vertex_data[10]),
        EdgeData(vertex_data[47], vertex_data[19]),

        EdgeData(vertex_data[48], vertex_data[11]),
        EdgeData(vertex_data[48], vertex_data[14]),
        EdgeData(vertex_data[49], vertex_data[11]),
        EdgeData(vertex_data[49], vertex_data[13])
    )

    num_windows = 12
    num_window_types = 1



"""
Energy Calculators
==================

#. :class:`.MMFFEnergy`
#. :class:`.UFFEnergy`
#. :class:`.MacroModelEnergy`
#. :class:`.MOPACEnergy`
#. :class:`.FormationEnergy`
#. :class:`.XTBEnergy`
#. :class:`.XTBFreeEnergy`


Energy calculators are objects which calculate the energy of molecules.
Each :class:`.EnergyCalculator` is initialized with some settings and
calculates the energy of a molecule with
:meth:`~.EnergyCalculator.get_energy`.

.. code-block:: python

    import stk

    # Energy calculators work with any Molecule objects, such as
    # BuildingBlock or ConstructedMolecule.
    mol1 = stk.BuildingBlock('[Br]CC[Br]', ['bromine'])

    chain = stk.polymer.Linear('A', [0], 12)
    mol2 = stk.ConstructedMolecule([mol1], chain)

    # Create the energy calculator.
    mmff = stk.MMFFEnergy()

    # Calculate energies of various molecules.
    mol1_energy = mmff.get_energy(mol1)
    mol2_energy = mmff.get_energy(mol2)

By default, calling :meth:`~.EnergyCalculator.get_energy` twice on the
same molecule will calculate the energy a second time. However, we can
use the `use_cache` option to prevent
recalculations when the same molecule is given to the same energy
calculator a second time

.. code-block:: python

    caching_mmff = MMFFEnergy(use_cache=True)
    # Calculate the energy the first time.
    energy1 = caching_mmff.get_energy(mol1)
    # The second time, the energy is returned directly from memory, a
    # second calculation is not run.
    energy2 = caching_mmff.get_energy(mol1)


.. _`adding energy calculators`:

Making New Energy Calculators
-----------------------------

New energy calculators can be made by simply making a class which
inherits the :class:`.EnergyCalculator` class. In addition to this,
the new class must define a :meth:`~.EnergyCalculator.get_energy`
method. The method must take 1 parameter, `mol`. The method will then
calculate and return the energy. There are no requirements regarding
how it should go about calculating the energy. New energy calculators
can be added into the :mod:`.energy_calculators` submodule or into a
new submodule.

"""

import rdkit.Chem.AllChem as rdkit
import logging
from functools import wraps
import subprocess as sp
import uuid
import os
from os.path import join
import shutil
from ...utilities import (
    is_valid_xtb_solvent,
    XTBInvalidSolventError,
    XTBExtractor
)


logger = logging.getLogger(__name__)


class EnergyError(Exception):
    """
    Indicates a failed energy calculation.

    """

    ...


def _add_cache_use(get_energy):
    """
    Makes :meth:`~EnergyCalculator.get_energy` use the cache.

    Decorates `get_energy` so that before running it checks if the
    :class:`.Molecule` has already had its energy calculated. If so,
    and :attr:`~EnergyCalculator._use_cache` is ``True``, then the
    energy value in the cache is returned.

    Parameters
    ----------
    get_energy : :class:`function`
        A function which is to have cache use added to it.

    Returns
    -------
    :class:`function`
        The decorated function.

    """

    @wraps(get_energy)
    def inner(self, mol):
        if self._use_cache and mol in self._cache:
            logger.info(f'Using cached energy value with {mol}.')
            return self._cache[mol]
        else:
            e = get_energy(self, mol)
            if self._use_cache:
                self._cache[mol] = e
            return e

    return inner


class EnergyCalculator:
    """
    Calculates the energy of molecules.

    """

    def __init__(self, use_cache=False):
        """
        Initialize a :class:`EnergyCalculator` instance.

        Parameters
        ----------
        use_cache : :class:`bool`, optional
            If ``True`` :meth:`get_energy` will not run twice on the
            same molecule, but will instead return the previously
            calculated value.

        """

        # Maps molecules to previously calculated energy values.
        self._cache = {}
        self._use_cache = use_cache

    def __init_subclass__(cls, **kwargs):
        cls.get_energy = _add_cache_use(cls.get_energy)
        return super().__init_subclass__(**kwargs)

    def get_energy(self, mol):
        """
        Calculate the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`float`
            The energy.

        Raises
        ------
        :class:`NotImplementedError`
            This is a virtual method and needs to be implemented in a
            subclass.

        """

        raise NotImplementedError()


class FormationEnergy(EnergyCalculator):
    """
    Calculates the formation energy of a molecule.

    Examples
    --------
    .. code-block:: python

        import stk

        # Make the molecules needed to calculate formation energy.
        water = stk.BuildingBlock('[H]O[H]')
        bb1 = stk.BuildingBlock('NCCCN', ['amine'])
        bb2 = stk.BuildingBlock('O=CCC=O', ['aldehyde'])

        chain = stk.polymer.Linear('AB', [0, 0], 6)
        polymer = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2],
            topology_graph=chain
        )

        # Make the energy calculator used to calculate energies.
        uff_energy = stk.UFFEnergy()

        # Make the formation energy calculator.
        formation = stk.FormationEnergy(
            energy_calculator=uff_energy,
            reactants=[bb1]*6 + [bb2]*6,
            products=[water]*polymer.bonds_made
        )

        # Calculate the formation energy.
        formation_energy = formation.get_energy(polymer)

    """

    def __init__(
        self,
        energy_calculator,
        reactants,
        products,
        use_cache=False
    ):
        """
        Initialize a :class:`FormationEnergy` instance.

        Parameters
        ----------
        energy_calculator : :class:`EnergyCalculator`
            The :class:`EnergyCalculator` used to calculate the energy
            of all reactant and product molecules.

        reactants : :class:`list` of :class:`.Molecule`
            The reactants. If there are multiples of the same reactant
            then it must appear multiple times in this :class:`list`.

        products : :class:`list` of :class:`.Molecule`
            The molecules which are produced as a result of the
            formation reaction. This :class:`list` must omit the
            :class:`.Molecule` passed to :meth:`get_energy`. If there
            are multiples of the same product, it must appear multiple
            times in this :class:`list`.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`get_energy` will not run twice on the
            same molecule, but will instead return the previously
            calculated value.

        """

        self._energy_calculator = energy_calculator
        self._reactants = reactants
        self._products = products
        super().__init__(use_cache=use_cache)

    def get_energy(self, mol):
        """
        Calculate the formation energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose formation energy is to be
            calculated.

        Returns
        -------
        :class:`float`
            The formation energy.

        """

        product_energy = sum(
            self._energy_calculator.get_energy(product)
            for product in self._products
        )
        product_energy += self._energy_calculator.get_energy(mol)

        reactant_energy = sum(
            self._energy_calculator.get_energy(reactant)
            for reactant in self._reactants
        )

        return product_energy - reactant_energy


class MMFFEnergy(EnergyCalculator):
    """
    Uses the MMFF force field to calculate energies.

    Examples
    --------
    .. code-block:: python

        import stk

        # Create a molecule whose energy we want to know.
        mol1 = stk.BuildingBlock('CCCNCCCN')

        # Create the energy calculator.
        mmff = stk.MMFFEnergy()

        # Calculate the energy.
        energy1 = mmff.get_energy(mol1)

    """

    def get_energy(self, mol):
        """
        Calculate the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`float`
            The energy.

        """

        rdkit_mol = mol.to_rdkit_mol()
        rdkit.SanitizeMol(rdkit_mol)
        rdkit.GetSSSR(rdkit_mol)
        ff = rdkit.MMFFGetMoleculeForceField(
            rdkit_mol,
            rdkit.MMFFGetMoleculeProperties(rdkit_mol)
        )
        return ff.CalcEnergy()


class UFFEnergy(EnergyCalculator):
    """
    Uses the UFF force field to calculate energies.

    Examples
    --------
    .. code-block:: python

        import stk

        # Create a molecule whose energy we want to know.
        mol1 = stk.BuildingBlock('CCCNCCCN')

        # Create the energy calculator.
        uff = stk.UFFEnergy()

        # Calculate the energy.
        energy1 = uff.get_energy(mol1)

    """

    def get_energy(self, mol):
        """
        Calculate the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`float`
            The energy.

        """

        rdkit_mol = mol.to_rdkit_mol()
        rdkit.SanitizeMol(rdkit_mol)
        # RingInfo needs to be initialized, else rdkit may raise an
        # error.
        rdkit.GetSSSR(rdkit_mol)
        ff = rdkit.UFFGetMoleculeForceField(rdkit_mol)
        return ff.CalcEnergy()


class XTBEnergy(EnergyCalculator):
    """
    Uses GFN-xTB [1]_ to calculate energy and other properties.

    By default, :meth:`get_energy` will extract other properties of the
    :class:`.Molecule` passed to :meth:`get_energy`, which
    will be saved in the attributes of :class:`.XTBEnergy`.

    Notes
    -----
    When running :meth:`get_energy`, this calculator changes the
    present working directory with :func:`os.chdir`. The original
    working directory will be restored even if an error is raised, so
    unless multi-threading is being used this implementation detail
    should not matter.

    If multi-threading is being used an error could occur if two
    different threads need to know about the current working directory
    as :class:`.XTBEnergy` can change it from under them.

    Note that this does not have any impact on multi-processing,
    which should always be safe.

    Attributes
    ----------
    total_energies : :class:`dict`
        :class:`dict` of the total energy of each :class:`.Molecule`
        passed to :meth:`get_energy`.

    homo_lumo_gaps : :class:`dict`
        :class:`dict` of the HOMO-LUMO gap of each :class:`.Molecule`
        passed to :meth:`get_energy`.

    fermi_levels : :class:`dict`
        :class:`dict` of the Fermi level of each :class:`.Molecule`
        passed to :meth:`get_energy`.

    homo_lumo_orbitals : :class:`dict`
        :class:`dict` of the HOMO-LUMO orbital properties of
        each :class:`.Molecule` passed to :meth:`get_energy`.

    qonly_dipole_moments : :class:`dict`
        :class:`dict` of the q only dipole moment of
        each :class:`.Molecule` passed to :meth:`get_energy`.

    full_dipole_moments : :class:`dict`
        :class:`dict` of the full dipole moment of
        each :class:`.Molecule` passed to :meth:`get_energy`.

    qonly_quadrupole_moments : :class:`dict`
        :class:`dict` of the q only quadrupole moment of
        each :class:`.Molecule` passed to :meth:`get_energy`.

    qdip_quadrupole_moments : :class:`dict`
        :class:`dict` of the q+dip quadrupole moment of
        each :class:`.Molecule` passed to :meth:`get_energy`.

    full_quadrupole_moments : :class:`dict`
        :class:`dict` of the full quadrupole moment of
        each :class:`.Molecule` passed to :meth:`get_energy`.

    total_free_energies : :class:`dict`
        :class:`dict` of the total free energy of
        each :class:`.Molecule` passed to :meth:`get_energy`.
        This is empty if :attr:`calculate_free_energy` is ``False``.

    frequencies : :class:`dict`
        :class:`dict` of the vibrational frequencies of
        each :class:`.Molecule` passed to :meth:`get_energy`.
        This is empty if `calculate_free_energy` was ``False``.

    Examples
    --------
    .. code-block:: python

        import stk

        bb1 = stk.BuildingBlock('NCCNCCN', ['amine'])
        bb2 = stk.BuildingBlock('O=CCCC=O', ['aldehyde'])
        polymer = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2],
            topology_graph=stk.polymer.Linear("AB", [0, 0], 3)
        )

        # Optimize the constructed molecule so that it has a
        # reasonable structure.
        optimizer = stk.OptimizerSequence(
            stk.ETKDG(),
            stk.XTB(xtb_path='/opt/gfnxtb/xtb', unlimited_memory=True)
        )
        optimizer.optimize(polymer)

        # Calculate energy using GFN-xTB.
        xtb = stk.XTBEnergy(
            xtb_path='/opt/gfnxtb/xtb',
            unlimited_memory=True
        )

        p_total_energy = xtb.get_energy(polymer)

        # Extract properties from the energy calculator for a given
        # molecule.
        homo_lumo_gap = xtb.homo_lumo_gaps[polymer]
        fermi_levels = xtb.fermi_levels[polymer]
        homo_lumo_orbitals = xtb.homo_lumo_orbitals[polymer]
        full_dipole_moments = xtb.full_dipole_moments[polymer]
        full_quadrupole_moments = xtb.full_quadrupole_moments[polymer]

        # The total energy can be extracted at any point from the
        # calculator.
        total_energy = xtb.total_energies[polymer]

    If `calculate_free_energy` is ``True``, xTB performs a
    numerical Hessian calculation and calculates the total free energy
    and vibrational frequencies of a molecule. It is recommended that a
    well optimized structure be used as input for these calculations

    .. code-block:: python

        # Optimize the constructed molecule so that it has a
        # reasonable structure.
        optimizer = stk.OptimizerSequence(
            stk.ETKDG(),
            stk.XTB(
                xtb_path='/opt/gfnxtb/xtb',
                unlimited_memory=True,
                opt_level='verytight'
            )
        )
        optimizer.optimize(polymer)

        # Calculate energy using GFN-xTB.
        xtb = stk.XTBEnergy(
            xtb_path='/opt/gfnxtb/xtb',
            unlimited_memory=True,
            calculate_free_energy=True
        )

        p_total_energy = xtb.get_energy(polymer)

        # Extract properties from the energy calculator for a given
        # molecule.
        p_total_free_energy = xtb.total_energies[polymer]
        p_frequencies = xtb.frequencies[polymer]

    References
    ----------
    .. [1] https://xtb-docs.readthedocs.io/en/latest/setup.html

    """
    def __init__(
        self,
        xtb_path,
        gfn_version=2,
        output_dir=None,
        num_cores=1,
        calculate_free_energy=False,
        electronic_temperature=300,
        solvent=None,
        solvent_grid='normal',
        charge=0,
        num_unpaired_electrons=0,
        unlimited_memory=False,
        use_cache=False
    ):
        """
        Initializes a :class:`XTBEnergy` instance.

        Parameters
        ----------
        xtb_path : :class:`str`
            The path to the xTB executable.

        gfn_version : :class:`int`, optional
            Parameterization of GFN to use in xTB.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/basics.html.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the calculation are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        num_cores : :class:`int`, optional
            The number of cores xTB should use.

        calculate_free_energy : :class:`bool`, optional
            Whether to calculate the total free energy and vibrational
            frequencies. Setting this to ``True`` can drastically
            increase calculation time and memory requirements.

        electronic_temperature : :class:`int`, optional
            Electronic temperature in Kelvin.

        solvent : :class:`str`, optional
            Solvent to use in GBSA implicit solvation method.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/gbsa.html.

        solvent_grid : :class:`str`, optional
            Grid level to use in SASA calculations for GBSA implicit
            solvent.
            Can be one of ``'normal'``, ``'tight'``, ``'verytight'``
            or ``'extreme'``.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/gbsa.html.

        charge : :class:`int`, optional
            Formal molecular charge.

        num_unpaired_electrons : :class:`int`, optional
            Number of unpaired electrons.

        unlimited_memory : :class: `bool`, optional
            If ``True`` :meth:`energy` will be run without constraints
            on the stack size. If memory issues are encountered, this
            should be ``True``, however this may raise issues on
            clusters.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`energy` will not run twice on the same
            molecule.

        """
        if solvent is not None:
            solvent = solvent.lower()
            if gfn_version == 0:
                raise XTBInvalidSolventError(
                    f'No solvent valid for version',
                    f' {gfn_version!r}.'
                )
            if not is_valid_xtb_solvent(gfn_version, solvent):
                raise XTBInvalidSolventError(
                    f'Solvent {solvent!r} is invalid for ',
                    f'version {gfn_version!r}.'
                )

        self._xtb_path = xtb_path
        self._gfn_version = str(gfn_version)
        self._output_dir = output_dir
        self._num_cores = str(num_cores)
        self._calculate_free_energy = calculate_free_energy
        self._electronic_temperature = str(electronic_temperature)
        self._solvent = solvent
        self._solvent_grid = solvent_grid
        self._charge = str(charge)
        self._num_unpaired_electrons = str(num_unpaired_electrons)
        self._unlimited_memory = unlimited_memory

        self.total_energies = {}
        self.homo_lumo_gaps = {}
        self.fermi_levels = {}
        self.homo_lumo_orbitals = {}
        self.qonly_dipole_moments = {}
        self.full_dipole_moments = {}
        self.qonly_quadrupole_moments = {}
        self.qdip_quadrupole_moments = {}
        self.full_quadrupole_moments = {}
        self.total_free_energies = {}
        self.frequencies = {}
        super().__init__(use_cache=use_cache)

    def _get_properties(self, mol, output_file):
        """
        Extracts properties from a GFN-xTB energy calculation.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy was calculated.

        output_file : :class: `str`
            Name of the output file with xTB results.

        Returns
        -------
        None : :class:`NoneType`

        """
        # Get properties from output_file.
        xtbext = XTBExtractor(output_file=output_file)

        self.total_energies[mol] = xtbext.total_energy
        self.homo_lumo_gaps[mol] = xtbext.homo_lumo_gap
        self.fermi_levels[mol] = xtbext.fermi_level
        self.homo_lumo_orbitals[mol] = xtbext.homo_lumo_occ
        self.qonly_dipole_moments[mol] = xtbext.qonly_dipole_moment
        self.full_dipole_moments[mol] = xtbext.full_dipole_moment
        self.qonly_quadrupole_moments[mol] = \
            xtbext.qonly_quadrupole_moment
        self.qdip_quadrupole_moments[mol] = \
            xtbext.qdip_quadrupole_moment
        self.full_quadrupole_moments[mol] = \
            xtbext.full_quadrupole_moment
        if self._calculate_free_energy:
            self.total_free_energies[mol] = xtbext.total_free_energy
            self.frequencies[mol] = xtbext.frequencies

    def _run_xtb(self, xyz, out_file):
        """
        Runs GFN-xTB.

        Parameters
        ----------
        xyz : :class:`str`
            The name of the input structure ``.xyz`` file.

        out_file : :class:`str`
            The name of output file with xTB results.

        Returns
        -------
        None : :class:`NoneType`

        """

        # Modify the memory limit.
        if self._unlimited_memory:
            memory = 'ulimit -s unlimited ;'
        else:
            memory = ''

        if self._solvent is not None:
            solvent = f'--gbsa {self._solvent} {self._solvent_grid}'
        else:
            solvent = ''

        if self._calculate_free_energy:
            calc_type = '--hess'
        else:
            calc_type = ''

        cmd = (
            f'{memory} {self._xtb_path} '
            f'{xyz} --gfn {self._gfn_version} '
            f'{calc_type} --parallel {self._num_cores} '
            f'--etemp {self._electronic_temperature} '
            f'{solvent} --chrg {self._charge} '
            f'--uhf {self._num_unpaired_electrons}'
        )

        with open(out_file, 'w') as f:
            # Note that sp.call will hold the program until completion
            # of the calculation.
            sp.call(
                cmd,
                stdin=sp.PIPE,
                stdout=f,
                stderr=sp.PIPE,
                # Shell is required to run complex arguments.
                shell=True
            )

    def get_energy(self, mol):
        """
        Calculate the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`float`
            The energy.

        """

        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)

        os.mkdir(output_dir)
        init_dir = os.getcwd()
        xyz = join(output_dir, 'input_structure.xyz')
        out_file = join(output_dir, 'energy.output')
        mol.write(xyz)

        try:
            os.chdir(output_dir)
            self._run_xtb(xyz=xyz, out_file=out_file)
        finally:
            os.chdir(init_dir)

        self._get_properties(
            mol=mol,
            output_file=out_file
        )
        return self.total_energies[mol]
